# dbbrowse.tcl --
#
# Package for browsing the databases
#
# Multiple browsers can be opened on the databases.
#
# Typically, a browser contains a window
# and a list. When text is entered in the top window,
# the list will be updated so that the text it displays
# be *after* the top window's text in alphabetic order.
#
#  1998 Serge Rosmorduc,
# quipe EC.art, Laboratoire LRIA, Universit de Paris 8
#
# 15 rue Catulienne
# 93236 St Denis Cedex
# FRANCE
#
# Tout droits rservs.
# 
# voyez le fichier ../LICENCE.txt pour plus d'informations
# See the ../LICENCEE.txt file for more information
#
# $Id$
#
#

#
# $Log$
#

# dbbrowse::VARS contains all data associated with browsers
# fields are written this way : VARS($w,FIELD)
# where w is the browser

# FUNCTION SUMMARY --

# known functions are : 
# dbbrowse::create -- creates a new browser
# dbbrowse::setfirst -- sets up the first data in the browser
# dbbrowse::getval -- returns the value associated with the current entry
# dbbrowse::setload -- sets the loading function called when a item is selected

# DATA SUMMARY -- 
# Fields of VARS : 
# head : upper tag of the entries we are interested in
# unnorm : function used to transform the items into their displayed form in the lists
# norm : from displayed form to list form 
# first : first item displayed in the list
# scan : the current scan item associated with the browser. See seshbase for more details
# callback : loading function called when an item is selected. The function takes the browser as argument
# idx : the database used as index for scanning. Default to idx (the index for the dico database)

#  set SESHLIB [pwd]
#  set env(SESHLIB) [pwd]
#  puts $env(SESHLIB)  
#  lappend auto_path $env(SESHLIB) \
#      [file join $env(SESHLIB) stooop]\
#      [file join $env(SESHLIB) scwoop]

# References used to be made to the dico class.
# Things should be more general.

package require seshutils
package require seshbase
package require tag
package require yellow
package require legendes


package provide dbbrowse 0.0

namespace eval dbbrowse {
    namespace export create
    namespace export setfirst
    namespace export setload
    namespace export getval

    
    variable VARS
    
# dbbrowse::create --
# 	creates a new browser
# Parameters:
#	w : the browser's name
# Options:
#	-head TAG : the upper tag of the entries the browser browses
#	-unnorm FUNC : the unnormalisation function. Takes the tag's value as argument
#	-norm FUNC : the normalisation function. Takes the tag's unnormalised value as argument
#	-idx: name of the base which is used as index
# Results:
#	the newly created browser
# Side Effects:
#	a browser is created
# Use:
#	PUBLIC
# Called by:

    proc create {w args} {
	variable VARS

	array set opt {
	    -head MTRI
	    -unnorm dbbrowse::identity
	    -norm dbbrowse::identity
	    -idx idx
	}
	
	set newargs {}
	
	foreach {a b} $args {
	    if [info exists opt($a)] {
		set opt($a) $b
	    } else {
		lappend newargs $a
		lappend newargs $b
	    }
	}

	set VARS($w,head) $opt(-head)
	set VARS($w,unnorm) $opt(-unnorm)
	set VARS($w,norm) $opt(-norm)
	set VARS($w,idx) $opt(-idx)

	frame $w
	eval entry $w.entr -textvariable dbbrowse::VARS($w,first)\
		$newargs

	yellow::create $w.entr [legendes::trans "Type here the beginning\n of the word you search"]
	
	eval "listbox $w.list -width 20 -height 10 -yscroll"\
		[list "dbbrowse::setbar $w"] {-takefocus 1}\
		$newargs

	#scrollbar $w.s -command "$w.list yview"
	scrollbar $w.s -command "dbbrowse::getbar $w"
	
	grid $w.entr -
	grid $w.list $w.s -sticky news

	grid columnconfigure $w 0 -weight 1
	grid rowconfigure $w 1 -weight 1

	bind $w.entr <Return> [list dbbrowse::init $w]
	bind $w <Destroy> {dbbrowse::free %W}
	bind $w.list <Return> [list dbbrowse::loadbr $w]
	bind $w.list <Double-Button-1> [list dbbrowse::loadbr $w]
	bind $w.list <Down> [list dbbrowse::down $w]

	init $w

	return $w
    }

    # dbbrowse::identity --
    # 	identity function used as default value for options which ask for a function
    # Parameters:
    #	x : x
    # Results:
    #	x
    # Use:
    #	CALLBACK
    # Called by:
    
    proc identity {x} {
	return $x
    }
    
    # dbbrowse::free --
    # 	frees the memory associated with the browser
    #   SHOULDN'T BE CALLED DIRECTLY. (called when the window $w is destroyed)
    # Parameters:
    #	w : the browser's name
    # Results:
    #	none
    # Side Effects:
    #	memory is freed
    # Use:
    #	CALLBACK
    # Called by:


    proc free {w} {
	variable VARS
	seshbase::close_scan $VARS($w,scan)
	foreach n [array names VAR $w,* ] {
	    unset VARS($n)
	}
    }
    
# dbbrowse::setbar --
# 	sets the scrollbar at a certain point
#       command called by the listbox
# Parameters:
# 	w: browser
# 	a: first point
# 	b: second point
# Results:
#	none
# Side Effects:
# 	the scrollbar moves, but the scrollbar's end
#       is not reached
# Use:
#	CALLBACK
# Called by:


    proc setbar {w a b} {
	#$w.s set $a [expr $a + 0.2]
	$w.s set $a  [expr $a + 0.2]
    }

# dbbrowse::getbar --
# 	get position data from the scrollbar
# Parameters:
#	w : the browser's name
# Options:
#	moveto A B : moves to position A
#	scroll A B : scrolls
# Results:
#	none
# Side Effects:
#	both scrollbar postion and data position are changed
# Use:
#	CALLBACK
# Called by:


    proc getbar {w args} {
	set com [lindex $args 0]
	switch -- $com {
	    moveto {
		set v [lindex $args 1]
		if {$v > 0.5} {
		    next $w
		}
	    }
	    scroll {
		set val [lindex $args 1]
		set unit [lindex $args 2]
		set yv [lindex [$w.list yview] 1]
		if {$yv >= 0.5} {
		    next $w
		}
	    }
	}
	eval $w.list yview $args
    }
 
    # dbbrowse::down --
    # 	command called when pressing down
    # Parameters:
    #	w : the browser's name
    # Options:
    # Results:
    #	none
    # Side Effects:
    #	if the scrollbar is in the lower 50%, load a new database chunk
    # Use:
    #	CALLBACK
    # Called by:
    
    
    proc down {w} {
	set yv [lindex [$w.list yview] 1]
	if {$yv >= 0.5} {
	    next $w
	}
    }

    # dbbrowse::getidx --
    # 	get the index associated with the current choice.
    # Parameters:
    #	w : the browser's name
    # Results:
    #	the index value of the current choice
    #	in the form {TAG NORMALIZED_VALUE}
    #	Should probably be used by the loading callback
    # Side Effects:
    #	none
    # Comment:
    # Use:
    #	PUBLIC
    # Called by

    proc getidx {w} {
	variable VARS
	set value [$w.list get active]
 	set head $VARS($w,head)
	return [tag::build $head [$VARS($w,norm) $value]]
    }
    

    # dbbrowse::getval --
    # 	get the database value associated with the current choice.
    #	the choice is seen as an index in the database specified by the 
    #	-idx option of create.
    #
    # Parameters:
    #	w : the browser's name
    # Results:
    #	the value associated with the current choice
    #	Should probably be used by the loading callback
    # Side Effects:
    #	none
    # Comment:
    # Use:
    #	PUBLIC
    # Called by


    proc getval {w} {
	variable VARS
	return [find_index $w [getidx $w]]
    }

    # dbbrowse::find_index --
    # Parameters:
    # Results:
    # Side Effects:
    # Comment:
    # Use:
    #	PRIVATE
    # Called by

    proc find_index {w value} {
	variable VARS
	set base $VARS($w,idx)
	seshbase::openb $base
	set result [seshbase::lread $base $value]
	seshbase::closeb $base
	return $result
    }

    # dbbrowse::setfirst --
    # 	set the first data in the list
    # Parameters:
    #	w: the browser's name
    #   first: a value for the first item.
    # Options:
    # Results:
    #	none
    # Side Effects:
    # 	the first displayed item becomes the first item in the index base with a 
    #	value greater or equal than first
    # Use:
    #	PUBLIC
    # Called by:
    
    proc setfirst {w first} {
	variable VARS
	set VARS($w,first) $first
	init $w
    }
    
    # dbbrowse::listfrom --
    # 	gets the NUM next items from the base
    #   (almost : in fact, either NUM, NUM - 1 if exclude is true, or 
    #    all items until the end)
    # Parameters:
    #	w: the browser's name
    #   num: the number of items to get
    # 	exclude: do we want the first item or not ?
    # Options:
    # Results:
    # 	none
    # Side Effects:
    #   the current position in the scan advances 
    #	the next items are added to the listbox
    # Use:
    #	PRIVATE
    # Called by:
    # 	listexpand, init

    
    proc listfrom {w num {exclude 0}} {
	variable VARS

	set t [seshbase::scanget $VARS($w,scan) $num]
	if $exclude {
	    set t [lrange $t 1 end]
	}

	foreach item $t {
	    $w.list insert end [$VARS($w,unnorm) [tag::value $item]]
	}
    }
    
    # dbbrowse::listexpand --
    # 	expands the list of displayed items
    # Parameters:
    #	w: the browser's name
    #	pos: where to expand the list of displayed items :
    #		end : at the end, start : at the beginning
    # Options:
    # Results:
    # 	none
    # Side Effects:
    #   the current position in the scan advances 
    #	the next items are added to the listbox
    # Use:
    #	PRIVATE
    # Called by:
    #   next

    
    proc listexpand {w pos} {
	switch $pos {
	    end {
		set h [expr 2 * [$w.list cget -height]]
		listfrom $w $h 1
	    }
	    start {
		error "I'm not able to do that yet"
	    }
	}
    }
    
    # dbbrowse::setload --
    # 	sets the loading function called when a item is selected
    # Parameters:
    #	w: the browser's name
    # 	fonc: the function (takes THE BROWSER as argument)
    # Options:
    # Results:
    # 	none
    # Side Effects:
    #   the loading function is set
    # Use:
    #	PUBLIC
    # Called by:

    
    proc setload {w fonc} {
	variable VARS
	set VARS($w,callback) $fonc
    }

    # dbbrowse::loadbr --
    # 	calls the loading function
    # Parameters:
    #	w: the browser's name
    # Options:
    # Results:
    # 	none
    # Side Effects:
    #   the loading function is called
    # Use:
    #	CALLBACK
    # Called by:

    
    proc loadbr {w} {
	variable VARS
	$VARS($w,callback) $w
    }

    # dbbrowse::init --
    # 	starts a new scan for the browser
    # Parameters:
    #	w: the browser's name
    # Options:
    # Results:
    # 	none
    # Side Effects:
    #   a new scan is started, and the listbox displays items starting
    #	with the FIRST variable
    # Use:
    #	CALLBACK, PRIVATE
    # Called by:
    #	create, setfirst

    proc init {w} {
	variable VARS
	set VARS($w,scan) [seshbase::open_scan \
			       $VARS($w,head)\
			       [tag::build $VARS($w,head) [$VARS($w,norm) $VARS($w,first)]]\
			       $VARS($w,idx)\
			      ]
	set h [expr 2 * [$w.list cget -height]]
	$w.list delete 0 end
	listfrom $w $h
    }

    # dbbrowse::next --
    # 	advances in the scan
    # Parameters:
    #	w: the browser's name
    # Options:
    # Results:
    # 	none
    # Side Effects:
    #   if the last element currently displayed in the listbox is 
    #   also the last element of the loaded list, then load a new chunk of data
    # Use:
    #	PRIVATE
    # Called by:
    #	getbar, down


    proc next {w} {
	variable VARS
	set last [lindex [$w.list yview] 1]
	if {$last >= 1} {
	    listexpand $w end
	}
	#$w.list yview scroll 1 units
    }
}

#  # essai
#  dbbrowse::create .w
#  grid .w -sticky news
#  grid rowconfigure . 0 -weight 1
#  grid columnconfigure . 0 -weight 1


