# dico.tcl --
#
#       
#	Dictionnary definition and interface for
#       the tksesh program. Tag oriented stuff with indexes
#       all over the place
#
# 
#
# Representation :
# A dictionary is a structure which contains :
#
# a main file and a list of indexes
# a max record number
#
# each record is indexed by a number (the max current number)
# there are indexes on a number of fields
# each index provides a LIST of the record numbers that contains the given
# value for the index
#
# each record in the main file has the following structure : 
# it is TAG oriented and built with lists.
# each TAG has the following form :
# {TAGNAME VALUE} 
# VALUE is mandatory. Its structure depends on the TAG. 
# an empty value is represented by {}.
#
#
#
# *************************************************************************
#
# In dumping, a tagon is replaced by "{TAGNAME" ... 
# and a tagoff is replaced by "}"
#
#  
#
#
# So, we should have now the following tags ::
# GRP : basic group container
# WRI : possible written form of the word
# TRAD : translation
# MTRI : translitteration du mot
# TRI  : texte en translitteration
# COM : a comment. The first area of the comment is the type of the comment, the next one 
#       is free text, containing both WRI, TRL, and TRI, texts. 
#       The text in COM is not currently indexed. It could be, but in his own index
# REF : a reference to a text
# AREF : reference to an article, or, generally, a written presentation
# GRA : Grammar
# HIE : hieroglyphic text.
# COP : coptic text.
# USE : usage
# PCDATA : Untagged text. 
#
#
# Each TAG is described by the following data :
#
# indentation
# above-line indentation
# start-string 
# end-string
# 
#
# Now, Some tags are legal in certain contexts and some other are not.
# Legal tags are :
# 
# CTXT       | LEGAL TAGS
# ___________|______________________________________________
# GRP        | WRI TRAD TRI COM REF AREF GRP USE GRA(Not PCDATA !)
# WRI        | Nothing. Special editor.
# HIE        | Nothing. Special editor.
# COM        | All
# TRAD       | PCDATA
# TRI        | PCDATA
# REF        | PCDATA
# AREF       | PCDATA
# USE        | HIE, TRI
# GRA        | Nothing for the moment
#
# dico.dtd is the reference for the final representation
#
# Structure for the database
#
# The records are in DICONAME.
# A special record is MAX, which contain the maximum id
#
# All indexes are in bases called DICONAME.TAG,
# where TAG is the indexed TAG.
#

package require legendes
package require ortho
package require seshbase
package require signes
package require seshutils
package require seshw
package require seshnormalize
package require Dbinter


package provide dico 0.0	;# This uses Tcl namespaces

namespace eval dico {

    namespace export create find_index fromtext loaddico
    namespace export tag_index

    catch {variable DICONAME [file join $env(SESHLIB) tmpdico]}

    # Arrays describing the dico windows. Each array describes a field. 
    # data for the array is set by using :
    # ARRAY(window)

    variable SUBWIN;		# nombre de fenetres hieroglyphiques dans une entree
    variable IDRECORD
    variable ISNEW;			# does the window display an old value ?
    variable OLDVAL;			# list of tags saved for the old window, for update purpose.


    variable ATOMIC_TAGS {};		# regular expression which recognise "atomic tags"

    
    variable LEGALTAGS 

    variable THE_TAGS
    array set THE_TAGS {}

    
    # args -- procedure to process an argument list
    # opts : name of the array that will contain the arguments
    # optlist : list of options and default value e.g. : {-dx 0 -dy 0}
    # args : the arguments to process
    
    proc args {opts optlist arglist} {
	upvar $opts o

	foreach {name value} $optlist {
	    set o($name) $value
	}

	foreach {name value} $arglist {
	    set code [catch {set tmp $o($name)} err]

	    if { $code != 0} {
		tk_dialog  .dialog  [legendes::trans "error"]\
		    [legendes::trans "A function received an unexpected argument: %s. This is a bug, please report" $name] \
		    "" 0 [legendes::trans "Ok"]
	    }
	    set o($name) $value
	}
    }

    # declare the various options for the tag.
    proc declare_tag {TAG args} {
	variable THE_TAGS
	variable LEGALTAGS
	args opts {
	    -begin "\n" -middle "" -end "\n" 
	    -conf "" -bconf "" -econf "" -save "" -load ""
	    -index 0
	    -bindings {}
	    -keybind {}
	    -insert_at_end 0
	    -normalize identity
	    -description "tag non dcrit"
	    -htmlsave ""
	    -htmlbegin ""
	    -htmlend ""
	    -latexsave ""
	    -latexbegin ""
	    -latexend ""
	} $args
	set THE_TAGS($TAG,begin) $opts(-begin)
	set THE_TAGS($TAG,middle) $opts(-middle)
	set THE_TAGS($TAG,end) $opts(-end)
	set THE_TAGS($TAG,conf) $opts(-conf)
	set THE_TAGS($TAG,bconf) $opts(-bconf)
	set THE_TAGS($TAG,econf) $opts(-econf)
	set THE_TAGS($TAG,save) $opts(-save)
	set THE_TAGS($TAG,load) $opts(-load)
	set THE_TAGS($TAG,index) $opts(-index)
	set THE_TAGS($TAG,bindings) $opts(-bindings)
	set THE_TAGS($TAG,keybind) $opts(-keybind)
	set THE_TAGS($TAG,normalize) $opts(-normalize)
	set THE_TAGS($TAG,insert_at_end) $opts(-insert_at_end)
	set THE_TAGS($TAG,description) $opts(-description)
	set THE_TAGS($TAG,htmlsave) $opts(-htmlsave)
	set THE_TAGS($TAG,htmlbegin) $opts(-htmlbegin)
	set THE_TAGS($TAG,htmlend) $opts(-htmlend)
	set THE_TAGS($TAG,latexsave) $opts(-latexsave)
	set THE_TAGS($TAG,latexbegin) $opts(-latexbegin)
	set THE_TAGS($TAG,latexend) $opts(-latexend)

        foreach {b com} $THE_TAGS($TAG,keybind) {
	    set THE_TAGS($TAG,bind,$b) $com
	}
    }



    # Declarations of TAGS.
    # To declare a tag, one must :
    # give it a name. A TAG name 
    # should be uppercase letters only.
    # No numerals !
    #
    # a) State its grammar (below)
    # b) describe it with declare_tag
    # d) insert creation menu where needed
    #    (for example :
    #	$w.menus.insert.m add command -label "usage" \
    #	    -command [list dico::insert_parenthesis $w USE])
    #
    # In the case of tags which create a new window, like hieroglyphs
    # One needs to 
    # 1) write a creation function. 
    #    The function will insert the window at the insertion point
    #    Its arguments should be {w {txt ""}}
    #    The name of the window should be $w.'s'TAG$SUBWIN($w). 
    #       e.g. $w.sWRI1
    #    SUBWIN should then be augmented  
    # 2) write the save code. The save code is a function that 
    #          takes $w and the base window as arguments, and
    #          should return a text. Note : to obtain the text window
    #          from the base window, you should add a '.t'
    #
    # 3) The save function and the load function should be passed 
    #    as arguments to declare_tag with the options -save and -load
    
    array set LEGALTAGS {
	GRP {GRP TRAD WRI MTRI COM DET REF AREF USE GRA AUT EXE EMP
	      VAR}
	WRI ""
	DET ""
	HIE ""
	COM {HIE MTRI TRI REF AREF GRP PCDATA USE EXE GRA COP}
	TRAD {PCDATA MOT QQUN QQCHOSE}
	TRI {PCDATA}
	COP {PCDATA}
	MTRI {PCDATA}
	REF {PCDATA}
	AREF {PCDATA}
	USE {HIE PCDATA MOT QQUN QQCHOSE MTRI}
	EXE {HIE PCDATA MOT QQUN QQCHOSE COP}
	MOT {}
	QQUN {}
	QQCHOSE {}
	GRA {}
	LATTER {}
	OK {}
	NK {}
	LE {}
	VAR {}
	AUT {PCDATA}
	EMP {OK NK LE LATTER ABRV}
	NOTE {TITLE SECTION}
	TITLE {PCDATA HIE TRI COP}
	SUBTITLE {PCDATA HIE TRI COP}
	SECTION {SUBTITLE PAR SECTION}
	PAR { PCDATA TRI HIE COP REF AREF}
	fin {}
	debut {}
    }

    
    # The catch below is due to pkg_mkindex being 
    # unable to both index and load packages. pkg_mkindex would try to execute these
    # commands, but without index (because it's busy building one), hence an error.

    catch {
	declare_tag WRI -save save_hiero -load insert_writing -index 1 \
	    -htmlsave save_hiero_html\
	    -latexsave save_hiero_latex\
	    -normalize norm_hiero\
	    -description [legendes::trans "written form of a word"]

	declare_tag DET -save save_hiero -load insert_determinative -index 1 \
	    -htmlsave save_hiero_html\
	    -latexsave save_hiero_latex\
	    -normalize norm_hiero\
	    -description [legendes::trans "Possible determinative for a word"]
	
	declare_tag HIE -save save_hiero -load insert_hiero \
	    -htmlsave save_hiero_html\
	    -latexsave save_hiero_latex\
	    -normalize norm_hiero\
	    -description [legendes::trans "Hieroglyphic Text"]
	
	declare_tag TRAD  \
		-begin [legendes::trans "\ntranslation : "]\
		-end ".\n"\
		-bconf "-foreground red"\
		-htmlbegin [concat "\n<dl>\n<dt>" [legendes::trans "\ntranslation : "] "<dd>"]\
		-htmlend "</dd></dl>\n"\
		-conf {-font {Times 14 italic}}\
		-index 1\
		-description [legendes::trans "translation"]
    
	declare_tag TRI \
		-begin " "\
		-end " "\
		-conf {-foreground blue -font {Transliteration 18}} \
		-htmlbegin {<FONT face="Transliteration">}\
		-htmlend {</FONT>}\
		-normalize seshnormalize::norm_translit\
		-description [legendes::trans "Text in transliteration"]
    
	declare_tag MTRI \
		-begin " "\
		-end " "\
		-conf {-foreground blue -font {Transliteration 18}} \
		-index 1\
		-htmlbegin {<FONT face="Transliteration">}\
		-htmlend {</FONT>}\
		-normalize seshnormalize::norm_translit\
		-description [legendes::trans "Transliteration of a word"]
	
	
	declare_tag COP \
		-begin " "\
		-end " "\
		-conf {-font {Coptic 18}}\
		-htmlbegin {<FONT face="Coptic">}\
		-htmlend {</FONT>}\
		-keybind {
	                 $ 
	                  }\
		-description [legendes::trans "Coptic text"]
	    
    
        declare_tag COM \
		-begin [legendes::trans "\ncomment : "]\
		-end ".\n"\
		-htmlbegin "<P>"\
		-htmlend "</P>\n\n"\
		-bconf "-foreground blue"\
		-conf {-font {Times 14} -lmargin1 [expr 10 * $level] -lmargin2 [expr 10 * $level]}\
		-description [legendes::trans "Free text comment"]
    

        declare_tag REF \
		-begin {[} -end {]} \
		-conf {-foreground darkblue -font {helvetica 13}}\
		-bindings {
	    {<Double-Button-1> {dico::see_ref %W}}
	    {<Enter> {dico::set_blue %W}}
	    {<Leave> {dico::unset_blue %W}}
	}\
	        -description [legendes::trans "Text reference"]

    declare_tag AREF \
	-begin {[} -end {]} \
	-conf {-foreground darkblue -font {helvetica 13}}\
        -description [legendes::trans "document reference"]
    
    declare_tag GRA \
	-begin "(" \
	-end ")" \
	-conf "-foreground blue"\
        -description [legendes::trans "grammar"]

    declare_tag  USE\
        -begin [legendes::trans "expr : "] \
	-bconf "-font {helvetica 13}"\
	-end "." \
	-htmlbegin {<P><FONT face="Transliteration">}\
	-htmlend "</FONT></P>\n\n"\
	-conf "-foreground blue -font {Transliteration 18}"\
        -description [legendes::trans "Expression"]
    
    declare_tag  EXE\
        -begin [legendes::trans "example : "] \
	-bconf "-font {helvetica 13}"\
	-end "." \
	-htmlbegin {<BLOCKQUOTE><FONT face="Transliteration">}\
	-htmlend "</Font></BLOCKQUOTE></P>\n\n"\
	-conf "-foreground blue -font {Transliteration 18}"\
       -description [legendes::trans "Example"]

    declare_tag MOT \
	-begin [legendes::trans "<word" ] \
	-end ">" \
	-insert_at_end 1\
	-htmlbegin "~~"\
	-conf "-foreground blue -font {helvetica 13}"\
	-description [legendes::trans "<WORD>"]
	    
    declare_tag EMP \
	-begin {[} -end {]} \
	-htmlbegin {[} -htmlend {]} \
	-conf "-foreground blue -font {helvetica 13}"\
	-description [legendes::trans "Domain and type of uses"]
	    
    declare_tag ABRV \
	-begin [legendes::trans ".abbrev"] -end {.} \
        -insert_at_end 1\
	-htmlbegin { <Small>ABR.</Small> }\
	-conf "-foreground blue -font {helvetica 13}"\
	-description [legendes::trans "<abbreviation>"]

    declare_tag LATTER \
	-insert_at_end 1\
	-htmlbegin { <Small>LATTER</Small> }\
	-begin [legendes::trans ".later"] -end {.} \
	-conf "-foreground blue -font {helvetica 13}"\
	-description [legendes::trans "<Later value>"]

    declare_tag VAR \
	-insert_at_end 1\
	-htmlbegin { <Small>VAR.</Small> }\
	-begin [legendes::trans ".var"] -end {.} \
	-conf "-foreground blue -font {helvetica 13}"\
	-description [legendes::trans "<Variant>"]
    
    declare_tag OK \
	-insert_at_end 1\
	-begin [legendes::trans ".Old Kingdom"] -end {.} \
	-htmlbegin [concat " <Small>" [legendes::trans ".Old Kingdom"] "</Small> "]\
	-conf "-foreground blue -font {helvetica 13}"\
	-description [legendes::trans "<Old Kingdom>"]

    declare_tag NK \
	-insert_at_end 1\
	-begin [legendes::trans ".New Kingdom"] -end {.} \
	-htmlbegin [concat " <Small>" [legendes::trans ".New Kingdom"] "</Small> "]\
	-conf "-foreground blue -font {helvetica 13}"\
	-description [legendes::trans "<New Kingdom>"]

    declare_tag LE \
	-insert_at_end 1\
	-begin [legendes::trans ".Late Egyptian"] -end {.} \
	-htmlbegin [concat " <Small>" [legendes::trans ".Late Egyptian"] "</Small> "]\
	-conf "-foreground blue -font {helvetica 13}"\
	-description [legendes::trans "<Late Egyptian>"]

    declare_tag QQCHOSE \
	-insert_at_end 1\
	-begin [legendes::trans "<Something"] \
	-end ">" \
	-htmlbegin [concat " <Small>" [legendes::trans "Something"] "</Small> "]\
	-conf "-foreground blue -font {helvetica 13}"\
	-description [legendes::trans "<Something>"]

    declare_tag QQUN \
	-insert_at_end 1\
	-begin [legendes::trans "<Someone"] \
	-end ">" \
	-htmlbegin [concat " <Small>" [legendes::trans "Someone"] "</Small> "]\
	-conf "-foreground blue -font {helvetica 13}"\
	-description [legendes::trans "<Someone>"]

    declare_tag AUT \
	-begin [legendes::trans "(Author :"] \
	-end ")" \
	-bconf "-font {helvetica}"\
	-conf "-foreground blue"\
	-description [legendes::trans "Signature"]

    declare_tag GRP \
	-begin "\n" \
	-end "\n" \
	-bconf "-foreground red -spacing1 10"\
	-econf "-foreground red"\
	-conf {-lmargin1 [expr 10 * $level] -lmargin2 [expr 10 * $level]}\
	 -description [legendes::trans "Group"]
	
    declare_tag NOTE \
	-begin "\n" \
	-end "\n" \
	-bconf "-foreground red -spacing1 10"\
	-econf "-foreground red"\
	-conf {-lmargin1 [expr 10 * $level] -lmargin2 [expr 10 * $level]}\
	-description [legendes::trans "Free text note"]

    declare_tag TITLE \
	    -index 1\
	    -begin "-" \
	    -end "-\n" \
	    -htmlbegin {<H1>}\
	    -htmlend "</H1>\n"\
	    -latexbegin {\begin{center}\huge } \
	    -latexend {\end{center}} \
	    -conf {-justify center -foreground blue -font {helvetica 18 bold}}\
	    -description [legendes::trans "note title"]

    declare_tag SUBTITLE \
	    -begin " " \
	    -end "\n" \
	    -htmlbegin {<H2>}\
	    -htmlend "</H2>\n"\
	    -latexbegin "\\section{" \
	    -latexend "}\n" \
	    -conf {-justify left -foreground blue -font {helvetica 16 bold}}\
	    -description [legendes::trans "section title"]


    declare_tag SECTION \
	    -begin "\n" \
	    -end "\n" \
	    -bconf "-foreground red -spacing1 10"\
	    -econf "-foreground red"\
	    -conf {-lmargin1 [expr 10 * $level] -lmargin2 [expr 10 * $level]}\
	    -description [legendes::trans "Section of text"]
	
    declare_tag PAR \
	    -begin " " \
	    -end "\n" \
	    -bconf "-foreground red -spacing1 10"\
	    -econf "-foreground red"\
	    -htmlbegin "<P>"\
	    -htmlend "</P>\n\n"\
	    -latexend "\n\n"\
	    -conf {-lmargin1 [expr 10 * $level] -lmargin2 [expr 10 * $level]}\
	    -description [legendes::trans "Paragraph"]
	
	variable ATOMIC_TAGS
	set ATOMIC_TAGS {TRAD_}
    }    


    # In mosts functions here, w means the window containing the whole data, not
    # the particular subwindow. For example, the text window is always w.t
    #
    # for function that operate on a particular subwindow, the window name is win


    # upper_tag returns the innermost tag of a specific kind at some place
    #proc upper_tag {w tag pos} {
    #	set ltags [$w.t tag names $pos]
    #	regsub -all -- "${tag}_(\[0-9\]+)|(\[^ \]+)|( )" $ltags {\1\3} ltags
    #	lappend ltags 0
    #	set ltags [lsort -integer  -decreasing $ltags]
    #	return "${tag}_[lindex $ltags 0]"
    #}
    
    proc tag_comp {t1 t2} {
	regexp {[0-9]+$} $t1 t1
	regexp {[0-9]+$} $t2 t2
	return [expr $t1 - $t2]
    }
    
    proc upper_tag {win {pos insert}} {
	variable VARS
    	set ltags [writingtags $win $pos]
	set w [base_window $win]
    	regsub -all -- "(debut|fin)_(\[^ \]+)" $ltags {} ltags
	regsub -all -- "((\[^ \]+)_(\[^ \]+))|(\[^ \]+)" $ltags {\1} ltags

	set ltags [lsort -command tag_comp $ltags]
	set VARS($w,uppertag) [lindex $ltags end]
	return $VARS($w,uppertag)
    }
    
    proc imbrication_level {w tag pos} {
	set ltags [writingtags $w.t]
	#regsub -all -- ".*_?${tag}_(\[0-9\]+)|(\[^ \]+)|( )" $ltags {\1\3} ltags
	regsub -all -- ".*_(\[0-9\]+)|(\[^ \]+)|( )" $ltags {\1\3} ltags
	lappend ltags 0
	set ltags [lsort -integer  -decreasing $ltags]
	return [expr [lindex $ltags 0] + 1]
    }


    # is_embedded returns 1 if the insertion point is inside a 
    # tagged area declared as "atomic"

    proc is_embedded {w} {
	variable ATOMIC_TAGS
	set ltags [writingtags $w.t]
	return [regexp -- $ATOMIC_TAGS $ltags toto]
    }


    # tagname returns the name part of a tag.
    proc tagname {tag} {
	regexp {^[^_]*} $tag result 
	return $result
    }

    proc is_not_legal {win tag} {
	variable LEGALTAGS
	variable VARS
	
	set w [base_window $win]
	set roottag $VARS($w,roottag)
	set ltags [concat ${roottag}_0 [writingtags $win]]
	# First, it is illegal to write anything in a 
	# begin or end tag :
	# (the general case works for these !)
	# if [regexp {^debut_|^fin_} $ltags] {
	#    return 1
	# }
	#
	# Then, the innermost tag must be compatible 
	# with the new tag :
	set ltags [lsort -command tag_comp $ltags]
	set lasttag [tagname [lindex $ltags end]]
	if {[lsearch -exact $LEGALTAGS($lasttag) $tag] == -1} {
	    return 1
	} else {
	    return 0
	}
    }

    # dico::insert_tag --
    #          Insert a tag at insertion point
    # Arguments:
    #	w	        name of the dico window
    #	tag	        tag to add
    #
    # Results:
    #        Returns the name of the created tag (the base tag name + some number)
    #        the markers for tag start and tag end are inserted
    #        the insertion point is set at the middle of the tag
    
    proc insert_tag {w tag} {
	variable THE_TAGS

	set l [imbrication_level $w $tag insert]
	set level $l
	set tagname ${tag}_$l
	set debname debut_${tag}_$l
	set finname fin_${tag}_$l

	eval $w.t tag configure $tagname $THE_TAGS($tag,conf)
	eval $w.t tag configure $debname $THE_TAGS($tag,bconf)
	eval $w.t tag configure $finname $THE_TAGS($tag,econf)

	foreach i $THE_TAGS($tag,bindings) {
	    $w.t tag bind $tagname [lindex $i 0] [lindex $i 1]
	}
	
	set context [writingtags $w.t]

	set start [$w.t index insert];	# beginning of the tagged region	
	$w.t insert $start $THE_TAGS($tag,begin) $context;	# beginning text
	set starttext [$w.t index insert]; # text start

	$w.t insert insert $THE_TAGS($tag,middle) $context
	set endtext [$w.t index insert]	
	$w.t insert insert $THE_TAGS($tag,end) $context; # insert at end of text
	$w.t tag add $debname $start $starttext; # tag the beginning part
	$w.t tag add $finname $endtext insert
	$w.t tag add $tagname $start insert
	set end [$w.t index insert]	
	if {$THE_TAGS($tag,insert_at_end)} {
	    $w.t mark set insert "$end"
	} else {
	    $w.t mark set insert "$starttext"
	}
	return $tagname
    }

    # insert_parenthesis is here to allow the insertion of
    # parenthesized constructs.
    #   w : the text window
    #   tag : base name of the tag
    # Options :
    #  -begin, -middle, -end : the text to be inserted at the start, middle, and end of the tag
    #  -conf  : configuration options for the tag.
    #          you can use the variable "level", which is the degree of embedding of the tag
    #  -bconf, -econf : special options for tag beginning and ending. Can be used to introduce vertical spaces
    #
    # returns the created tag, and {} if it wasn't created

    proc insert_parenthesis {w tag} {
	if {[is_not_legal $w.t $tag]} {
	    tk_dialog  .dialog  [legendes::trans "Beware"] \
		[legendes::trans "You can't insert %s here" $tag] \
		"" 0 [legendes::trans "Ok"]
	    return {}
	}
	insert_tag $w $tag
    }

    proc insert_hieroglyphic_tag {w tag {txt ""}} {
	variable SUBWIN
	incr SUBWIN($w)
	set s $w.s$tag$SUBWIN($w)
	::seshw::create_passive $s
	$w.t window create insert -window $s
	if { $txt != "" } {
	    ::seshw::set_text $s $txt
	    ::seshw::autosize $s
	}
	bind $s <Double-Button-1> [list dico::edit_hieroglyphs $w $s]
    }
 
    proc edit_hieroglyphs {w subwin} {
	set ed .dicohieredit
	toplevel $ed
	wm title $ed [legendes::trans "hieroglyphic word editor"]
	seshw::create_editor $ed.s -notitle
	grid $ed.s -
	$ed.s.s insert 0.0 [$subwin get 0.0 end]
	button $ed.ok -text [legendes::trans "Ok"]\
		-command [list dico::sethierofromedit $subwin $ed]
	button $ed.cancel -text [legendes::trans "Cancel"]\
		-command [list destroy $ed]
	grid $ed.ok $ed.cancel
    }
    
    proc sethierofromedit {s orig} {
	seshw::set_text $s [$orig.s.s get 0.0 end]
	seshw::autosize $s
	destroy $orig
    }

    proc insert_writing {w {txt ""}} {
	insert_hieroglyphic_tag $w WRI $txt
    }
    
    proc insert_hiero {w {txt ""}} {
	insert_hieroglyphic_tag $w HIE $txt
    }

    proc insert_determinative {w {txt ""}} {
	insert_hieroglyphic_tag $w DET $txt
    }

    proc save_hiero {w hie} {
	return [$hie get 0.0 end]
    }

    proc save_hiero_html {base dir num window} {
	set fname img$num.xbm
	set f [open [file join $dir $fname] w]
	puts $f [$window undocumented4 -xbm]
	close $f
	return [format {<IMG src="%s">} $fname]
    }

    proc save_hiero_latex {base num window} {
	return [format {\begin{hieroglyph}
	    %s
	    \end{hieroglyph}} [$window get 0.0 end]]
    }

    proc create {w {type GRP}} {
	variable SUBWIN 
	variable GRAPHIES;		# les graphies principales 
	variable ISNEW
	variable OLD_TAGS
	
	# New, better system
	variable VARS;		# currently only hold $w,roottag
	set VARS($w,roottag) $type
	set VARS($w,uppertag) $type
	
	# creates a dictionary window
	
	# menus 
	frame $w
	frame $w.menus
	menubutton $w.menus.file -text [legendes::trans "Card"] -menu $w.menus.file.m	
	menubutton $w.menus.edit -text [legendes::trans "Edit"] -menu $w.menus.edit.m
	#menubutton $w.menus.insert -text "insert" -menu $w.menus.insert.m
	#menu $w.menus.insert.m
	menu $w.menus.file.m
	menu $w.menus.edit.m
	label $w.menus.tag -textvariable dico::VARS($w,uppertag)
	#pack $w.menus.file $w.menus.edit $w.menus.insert -side left
	pack $w.menus.file $w.menus.edit -side left
	pack $w.menus.tag -side right

	#variable ATOMIC_TAGS
	#set ATOMIC_TAGS {TRAD_}

	$w.menus.file.m add command -label [legendes::trans "New"] -command "dico::new $w"
	$w.menus.file.m add command -label [legendes::trans "Save"] -command "dico::save $w"
	$w.menus.file.m add command -label [legendes::trans "Save as HTML file"] -command "dico::save_as_html $w"
	$w.menus.file.m add command -label [legendes::trans "Save as LaTeX file"] -command "dico::save_as_latex $w"
	$w.menus.file.m add command -label [legendes::trans "Close"] -command "destroy $w"
#	$w.menus.file.m add command -label [legendes::trans "Load"] -command "dico::loaddico"

	$w.menus.edit.m add command -label [legendes::trans "Select current group"] \
	    -command [list dico::select_tag $w]

	$w.menus.edit.m add command -label [legendes::trans "Delete selected group"] \
	    -command [list $w.t delete sel.first sel.last]

	$w.menus.edit.m add command -label [legendes::trans "copy selected group"] \
	    -command [list dico::copy $w]
	
	$w.menus.edit.m add command -label [legendes::trans "Paste"]\
	    -command [list dico::paste $w]

	$w.menus.edit.m add command -label [legendes::trans "Paste Text"]\
	    -command [list dico::pastetext $w]

	frame $w.entete 

	#text window and scollbars
	text $w.t \
	    -yscrollcommand "$w.y set" \
	    -xscrollcommand "$w.x set"\
	    -wrap word\
	    -font {times 14}
	scrollbar $w.x -command "$w.t xview" -orient horizontal
	scrollbar $w.y -command "$w.t yview" -orient vertical
	bindtags $w.t [list $w.t HierarchicText all] 

	# grid
	grid $w.menus -sticky ew
	grid $w.t $w.y -sticky news
	grid $w.x -sticky ew
	grid rowconfigure $w 1 -weight 1
	grid columnconfigure $w 0 -weight 1
	
	# useful stuff for the text window :(TO SUPPRESS ?)
	$w.t tag configure debut -foreground red -spacing1 10
	$w.t tag configure fin -foreground red -spacing3 10
	$w.t tag configure translit -foreground blue -font {times 14 italic}

	$w.t tag configure blue -background lightblue
	set SUBWIN($w) 0
	set ISNEW($w) 1
	set OLD_TAGS($w) {}
    }

    proc base_window {win} {
	regsub {\.[^.]*$} $win {} w
	return $w
    }

    # returns a tag name for indexing
    # (values are normalised)

    proc tag_index {tag value} {
	variable THE_TAGS
	set v [$THE_TAGS($tag,normalize) $value]
	return [list $tag $v]
    }

    proc find_index {tag value} {
	variable THE_TAGS
	seshbase::openb idx
	set result [seshbase::lread idx \
			[tag_index $tag $value]]
	seshbase::closeb idx
	return $result
    }

    proc load {w id} {
	variable ISNEW;
	variable IDRECORD;
	variable OLD_TAGS

	seshbase::openb dico
	set txt [seshbase::readb dico $id]
	seshbase::closeb dico
	
	set ISNEW($w) 0	
	set IDRECORD($w) $id
	set OLD_TAGS($w) [list_tags $txt]
	fromtext $w $txt
    }

    proc new {w} {
	variable ISNEW;	
	$w.t delete 0.0 end
	set ISNEW($w) 1
    }



    proc identity {x} {
	return $x
    }

    # dico::list_tags --
    #          returns the tags to be indexed from a dictionary entry
    # Arguments:
    #   txt : text description of the entry
    #
    # Results:
    #        A list of elements {attributes value}, 
    #        describing the data to be saved

    proc list_tags {txt} {
	set tmp [save_aux $txt]
	return [uniq $tmp]
    }
    
    # dico::save_aux --
    #          returns the tags to be indexed from a dictionary entry
    # Arguments:
    #   txt : text description of the entry
    #
    # Results:
    #        A list of elements {attributes value}, 
    #        describing the data to be saved

    proc save_aux {txt} {
	variable THE_TAGS
	set result ""
	if {[is_tag $txt]} {
	    set tag_name [lindex $txt 0]
	    set tag_data [lindex $txt 1]

	    if {$THE_TAGS($tag_name,index)} {
		lappend result [tag_index $tag_name $tag_data]
	    } else {
		foreach i $tag_data {
		    set result [concat $result [save_aux $i]]
		}
	    }
	}
	return $result
    }

    
    #
    # get_orth --
    # sends back a list of couples :
    #   {trans orth}
    # where trans is a transliteration found in the definition
    # and orth is one of its orthographies.
    
    # Build(txt, previous_t, previous_h)
    # Algorithm : we start with an environment
    # which can contain a list of orthographies and/or of transliterations,
    #
    #
    # For the current group :
    #    g:= list of all subgroups.
    #    t:= list of transliterations in current group.
    #    h:= list of orthographies in current group.
    #
    #    if t = {} then t0:= previous_t else t0:= t
    #    if h = {} then h0:= previous_h else h0:= h
    #    
    #    result:= {}
    #
    #    if t != {}  or h != {}
    #      result:= {t0 h0}
    #
    #    foreach gr in g 
    #      b1:= Build(gr, t0, h0, result)
    #      result:= result + b1
    #
    #    return result

    
    proc find_translit {txt} {
	set res {}
	foreach elt $txt {
	    if [regexp {^MTRI *{(.*)}$} $elt dum tra] {
		lappend res $tra
	    }
	}
	return $res
    }

    proc find_hiero {txt} {
	set res {}
	foreach elt $txt {
	    if [regexp {^WRI *{(.*)}$} $elt dum tra] {
		lappend res $tra
	    }
	}
	return $res
    }

    proc find_subgroups {txt} {
	set res {}
	foreach elt $txt {
	    if [regexp {^GRP *{(.*)}$} $elt dum tra] {
		lappend res $tra
	    }
	}
	return $res
    }

    proc empty_list {l} {
	return [expr [string length $l] == 0]
    }

    proc build_orth {txt prevt prevh} {
	set t [find_translit $txt]
	set h [find_hiero $txt]
	set g [find_subgroups $txt]
	
	if [empty_list $t] {
	    set t0 $prevt
	} else {
	    set t0 $t
	}

	if [empty_list $h] {
	    set h0 $prevh
	} else {
	    set h0 $h
	}
	
	if {! [empty_list $t] || ! [empty_list $h]} {
	    set result [list [list $t0 $h0]]
	} else {
	    set result {}
	}
	
	foreach gr $g {
	    set b1 [build_orth $gr $t0 $h0]
	    set result [concat $result $b1]
	}
	return $result
    }

    # get_orth returns a list of the form :
    # trans orthographies trans orthographies ...
    
    proc get_orth {txt} {
	array set tmp {}
	set res [build_orth [list $txt] {} {}]
	foreach sub $res {
	    set trs [lindex $sub 0]
	    set h [lindex $sub 1]
	    foreach tr $trs {
		set v {}
		if [info exists tmp($tr)] {
		    set v $tmp($tr)
		}
		set tmp($tr) [concat $v $h]
	    } 
	}
	set result [array get tmp]
	return $result
    }

    proc save_orthography {txt} {
	foreach {t h} [get_orth $txt] {
	    if {! [empty_list $t] && ! [empty_list h]} {
		ortho::write $t $h
	    }
	}
    }

    # dico::save --
    #          save the content of window w to the disk
    # Arguments:
    #   w : window to save
    #
    # Results:
    #   no result returned.
    #   the data is saved
    #   the indexes are set to their proper values :
    #     * new data is entered
    #     * old data is cleared
    
    proc save {w} {
	variable ISNEW;
	variable IDRECORD;
	variable OLD_TAGS
	
	set txt [astext $w 0.0 end 1]
	# If the record is new, or doesn't belong to its saver, get a number
	seshbase::openb dico
	if {$ISNEW($w) || [seshbase::isnotmine $IDRECORD($w)]} {
	    set MAX [seshbase::get_max]
	    set IDRECORD($w) $MAX
	    set ISNEW($w) 0
	    set OLD_TAGS($w) {}
	}
	# save the record to disk
	seshbase::writeb dico $IDRECORD($w) $txt
	
	# open the index :
	seshbase::openb idx

	# compare the old tags and the new ones
	set newtags [list_tags $txt]

	# for all which are in "old - new"
	# suppress the corresponding reference in the index
	foreach i [set_less $OLD_TAGS($w) $newtags] {
	    set cur [seshbase::read1 idx $i]
	    set val [set_less $cur $IDRECORD($w)]
	    seshbase::writeb idx $i $val
	}

	# for all which are in "new - old"
	# add them in the index
	foreach i [set_less $newtags $OLD_TAGS($w)] {
	    set cur [seshbase::read1 idx $i]
	    if {[lsearch -exact $cur $IDRECORD($w)] == -1} { 
		lappend cur $IDRECORD($w)
		seshbase::writeb idx $i $cur
	    }
	}

	save_orthography $txt
	seshbase::closeb idx
	seshbase::closeb dico
    }

    proc protect {txt} {
	regsub -all -- {%}  $txt {%%} txt
	regsub -all --  \{  $txt {%par} txt
	regsub -all -- \}  $txt {%rap} txt
	regsub -all -- \"  $txt {%quo} txt
	#regsub -all -- { +}  $txt { %_ } txt

	return $txt
    }

    proc deprotect {txt} {
	regsub -all -- {%%} $txt {} txt
	regsub -all %par $txt \{ txt
	regsub -all %rap $txt \} txt
	regsub -all %quo $txt \" txt
	#regsub -all %_ $txt { } txt
	regsub -all -- {}  $txt {%} txt
	return $txt
    }


    
    # The dump command of the text widget sends back 
    # a number of irrelevant text separations. We 
    # deal with these.
    
    proc dumptext {w {debut 0.0} {fin end}} {
	set txt [$w.t dump -text -tag -window $debut $fin]
	set last {}
	foreach {k v i} $txt {
	    switch --  $k {
		text {
		    append last $v
		}
		default {
		    if {[string compare $last ""] != 0} {
			lappend res text
			lappend res $last
			lappend res 0
			set last {}
		    }
		    lappend res $k
		    lappend res $v
		    lappend res $i
		}
	    }
	}
	if {[string compare $last ""] != 0} {
	    lappend res text
	    lappend res $last
	    lappend res 0
	    set last {}
	}
	return $res
    }
    
    proc astext {w {debut 0.0} {fin end} {all 0}} {
 	variable THE_TAGS
	variable VARS
 	#Save the window -- a text window
 	#First, get the text for the whole window.
	
 	# save is used to mark text when we don't want to save it
	
 	set save 1
	set result "" 
	set espace ""

	set txt [dumptext $w $debut $fin]
	foreach {key value index} $txt {
	    switch -exact -- $key {
		text {
		    if {$save} {
			append result $espace
			set espace " "
			append result [protect $value]
		    }
		}
		tagon {
		    switch -glob -- $value {
			debut_* {
			    regexp {^debut_([^_]*)_} $value tmp tagname
			    append result "$espace\{$tagname \{"
			    set espace ""
			    set save 0
			}
			fin* {
			    set save 0
			}
			default {}
		    }
		}
		tagoff {
		    switch -glob -- $value {
			debut_* {
			    set save 1
			}
			fin* {
			    set save 1
			    append result "\}\}"
			    set espace " "
			}
			default {}
		    }
		}
		window {
		    regexp {\.s([^0-9]+)[0-9]+} $value dum tagname
		    append result $espace
		    set espace " "
		    append result [list [maketag \
					     $tagname \
	    [protect [$THE_TAGS($tagname,save) $w $value]]]]
		} 
		default {}
	    }
	}
	
	# Due to the tagging system and to the dump command, we can :
	#    * have extra tagoff command at the beginning of the text
	#    * miss tagoffs at the end
	# So, we should :
	# \{(1) suppress any trailing '\}'
	# \{(2) balance the string with the necessary '\}'
	# thanks to protect, it should work.
	
	if {! $all } {
	    regsub "^\}* *" $result "" result
	    regsub -all "\[^\}\]" $result "" a
	    regsub -all "\[^\{\]" $result "" b
	    set n [expr [string length $b] - [string length $a]]
	    for {set i 0} {$i < $n} {incr i} {
		append result "\}"
	    }
	}
	
	set result [maketag $VARS($w,roottag)  $result]
	return $result
    }

    # the following function will dump a page 
    # as a html document. 
    # When tags are provided for a window, they will be generated.
    # hieroglyphs will be dumped as graphic files.

    proc save_as_html {w} {
	set fic [tk_getSaveFile]
	if {[string compare $fic ""] != 0} {
	    dump_as_html $w $fic
	}
    }

    proc save_as_latex {w} {
	set fic [tk_getSaveFile]
	if {[string compare $fic ""] != 0} {
	    dump_as_latex $w $fic
	}
    }

    proc dump_as_html {w dirname} {
	variable THE_TAGS
	variable VARS
	set imgnum 0
 	#Save the window -- a text window
 	#First, get the text for the whole window.
	
 	# save is used to mark text when we don't want to save it
	
	file mkdir $dirname

 	set save 1
	set result "<HTML>\n<BODY>\n"
	set espace ""

	set txt [dumptext $w 1.0 end]
	foreach {key value index} $txt {
	    switch -exact -- $key {
		text {
		    if {$save} {
			append result $espace
			set espace " "
			append result $value
		    }
		}
		tagon {
		    switch -glob -- $value {
			debut_* {
			    regexp {^debut_([^_]*)_} $value tmp tagname
			    if [info exists THE_TAGS($tagname,htmlbegin)] {
				append result $THE_TAGS($tagname,htmlbegin)
			    }
			    set espace ""
			    set save 0
			}
			fin* {
			    set save 0
			}
			default {}
		    }
		}
		tagoff {
		    switch -glob -- $value {
			debut_* {
			    set save 1
			}
			fin* {
			    set save 1
			    regexp {^fin_([^_]*)_} $value tmp tagname
			    if [info exists THE_TAGS($tagname,htmlend)] {
				append result $THE_TAGS($tagname,htmlend)
			    }
			    set espace " "
			}
			default {}
		    }
		}
		window {
		    regexp {\.s([^0-9]+)[0-9]+} $value dum tagname
		    set fonc $THE_TAGS($tagname,htmlsave)
		    if {[string compare $fonc ""] != 0} {
			incr imgnum
			append result [$fonc $w $dirname $imgnum $value]
			set espace " "
		    }
		} 
		default {}
	    }
	}
	append result "</BODY>\n</HTML>\n"
	set f [open [file join $dirname index.html] w]
	puts $f $result
	close $f
    }

    proc dump_as_latex {w file} {
	variable THE_TAGS
	variable VARS
	set imgnum 0
 	#Save the window -- a text window
 	#First, get the text for the whole window.
	
 	# save is used to mark text when we don't want to save it

 	set save 1
	set result {
\documentclass{article}
\usepackage{egypto}
\usepackage{hiero}
\usepackage[latin1]{inputenc}
\usepackage[T1]{fontenc}

\begin{document}
	}
	
	set espace ""

	set txt [dumptext $w 1.0 end]
	foreach {key value index} $txt {
	    switch -exact -- $key {
		text {
		    if {$save} {
			append result $espace
			set espace " "
			append result $value
		    }
		}
		tagon {
		    switch -glob -- $value {
			debut_* {
			    regexp {^debut_([^_]*)_} $value tmp tagname
			    if [info exists THE_TAGS($tagname,latexbegin)] {
				append result $THE_TAGS($tagname,latexbegin)
			    }
			    set espace ""
			    set save 0
			}
			fin* {
			    set save 0
			}
			default {}
		    }
		}
		tagoff {
		    switch -glob -- $value {
			debut_* {
			    set save 1
			}
			fin* {
			    set save 1
			    regexp {^fin_([^_]*)_} $value tmp tagname
			    if [info exists THE_TAGS($tagname,latexend)] {
				append result $THE_TAGS($tagname,latexend)
			    }
			    set espace " "
			}
			default {}
		    }
		}
		window {
		    regexp {\.s([^0-9]+)[0-9]+} $value dum tagname
		    set fonc $THE_TAGS($tagname,latexsave)
		    if {[string compare $fonc ""] != 0} {
			incr imgnum
			append result [$fonc $w $imgnum $value]
			set espace " "
		    }
		} 
		default {}
	    }
	}
	append result {\end{document}}
	set f [open $file w]
	puts $f $result
	close $f
    }


    ## Fill a dictionary entry widget according to the text :
    
    # fromtext returns 1 if the text inserted requires a white space after it,
    # and else 0

    proc fromtext {w txt {all 1}} {
	variable THE_TAGS
	if {$all} {
	    $w.t delete 0.0 end 
	}
	
	set space ""
	set wantspace 0
	
	if {[is_tag $txt]} {
	    set tag_name [lindex $txt 0]
	    set tag_data [lindex $txt 1]
	    
	    if { $THE_TAGS($tag_name,load) != "" } {
		$THE_TAGS($tag_name,load) $w [deprotect $tag_data]
		set wantspace 0
	    } else {
		if {! $all } {
		    set currenttag [insert_tag $w $tag_name]
		}
		foreach i $tag_data {
		    $w.t insert insert $space
		    if [fromtext $w $i 0] {
			set space " "
			set wantspace 1
		    } else {
			set space ""
			set wantspace 0
		    }
		}
		if {! $all } {
		    set end [lindex [$w.t tag prevrange $currenttag insert] 1]
		    $w.t mark set insert $end
		}
		set wantspace 0
	    }
	} else {
	    $w.t insert insert [deprotect $txt]
	    set wantspace 1
	}
	return $wantspace
    }

    proc is_tag {txt} {
	return [expr [llength $txt] != 1]
    }

    # dico::maketag --
    #          build a correct tag
    #
    # Arguments:
    #   tag : name of the tag
    #   val : content
    #
    # Results:
    #   a correct tag.
    #   Note that if the tag value is one element long,
    #   there will be accolades anyway.
    #   In another life, I'll make PCDATA a first class TAG

    proc maketag {TAG val} {
	if { $val == "" } {
	    return [list $TAG {}]
	} else {
	    if {[llength $val] > 1} {
		return [list $TAG $val]
	    } else {
 		return "$TAG {$val}"
	    }
	}
    }

    proc select_tag {w} {
	# First, we find the previous begin tag for the current tag
	# next, we find the end tag. And voila, we have won
	set tag [upper_tag $w.t]
	set txt [$w.t tag prevrange debut_$tag  insert]
	#regexp {$([^_]+)_} $tag tagname
	if { $txt == "" } {
	    return 0
	}
	set debut [lindex $txt 0]
	set txt [$w.t tag nextrange fin_$tag  insert]
	if { $txt == "" } {
	    return 0
	}
	set fin [lindex $txt 1]
	if [$w.t compare $fin < insert] {
	    return 0
	}
	$w.t tag remove sel 0.0 end
	$w.t tag add sel $debut $fin
	$w.t tag lower sel
    }

    ################################################################################
    #
    # Special bindings for Text widget containing hierarchic tags.
    # The idea is that text under these tags are 
    # a) not easy to delete 
    # b) impossible to edit. In fact, they are usually \n
    bind HierarchicText <Control-Tab> {
    focus [tk_focusNext %W]
    }
    bind HierarchicText <Control-Shift-Tab> {
    focus [tk_focusPrev %W]
    }
    bind HierarchicText <Delete> {dico::textdelete %W; break}
    bind HierarchicText <BackSpace> {dico::textbackspace %W; break}
    bind HierarchicText <KeyPress> {dico::keyinsert %W %A; break}
    bind HierarchicText <Return> {dico::keyinsert  %W "\n"; break}
    bind HierarchicText <F10> {expr 1}
    bind HierarchicText <Button-3> {
	dico::tagchoice %W %X %Y
    }
    bind HierarchicText <Menu> {
	dico::tagchoice %W %X %Y
    }
    bind HierarchicText <Key-F2> {
	dico::tagchoice %W %X %Y
    }
    bind HierarchicText <Button-1> {
	focus %W
	%W mark set insert @%x,%y
	%W see insert
	dico::upper_tag %W
    }
    

    bind HierarchicText <Left> {dico::textleft %W; break}
    bind HierarchicText <Right> {dico::textright %W; break}
    bind HierarchicText <Up> {dico::textup %W; break}
    bind HierarchicText <Down> {dico::textdown %W; break}

    bind HierarchicText <<Paste>> {dico::winpastetext %W; break}


    proc textprotected {win pos} {
	return [expr [lsearch -regexp [$win tag names $pos] {debut_.*|fin_.*}] != -1]
    }

    proc textdelete {win} {
	if {! [textprotected $win insert] && [$win compare "insert + 1 chars" != end]} {
	    $win delete insert
	}
    }

    proc textbackspace {win} {
	if {! [textprotected $win "insert - 1 chars"]} {
	    $win delete "insert - 1 chars"
	}
    }

    # It is not possible to insert a character IN a "debut" or "fin" tag.
    # This means it is not possible to insert a character if both the current AND
    # the previous characters are in the same debut or fin.
    # That is, TAGS(insert) INTER TAGS(insert - 1 chars) does not contain 
    # a debut or fin tag
    
    # The cursor is "in" a tag iff
    # a) both sides of the cursor have the tag
    # b) the left side is NOT "fin_TAG" or the right side is
    #    not "debut_TAG", in which case we are between two area marked with the tag
    proc writingtags {win {pos insert}} {
	set position [$win index $pos]
	if {[$win compare 1.0 == $position]} {
	    return {}
	}
	set l2 [$win tag names $position]
	set l1 [$win tag names "$position - 1 chars"]
	# these lists will be short, so :
	set l {}
	foreach i  $l1 {
	    if {[lsearch -exact $l2 $i] != -1} {
		if {
		    [lsearch -exact $l1 fin_$i] == -1 ||
		    [lsearch -exact $l2 debut_$i] == -1
		} {
		    lappend l $i
		}
	    }
	}
	regsub sel $l {} l 
	return $l
    }
    
    # Is it possible to insert PCDATA at the current position ?
    proc textwritable {win} {
	return [expr ! [is_not_legal $win PCDATA]]
    }


    proc keyvalue {win k} {
	variable THE_TAGS
	set w [base_window $win]
	set t [tagname [upper_tag $win]]
	if [info exists THE_TAGS($t,bind,$k)] {
	    return $THE_TAGS($t,bind,$k)
	} else {
	    return $k
	}
    }

    proc keyinsert {win k} {
	textinsert $win [keyvalue $win $k]
    }

    proc textinsert {win k} {
	if [textwritable $win] {
	    $win insert insert $k
	    $win see insert
	    upper_tag $win
	}
    }

    proc copy {w} {
	clipboard clear
	set txt [astext $w sel.first sel.last]
	set txt [lindex [lindex $txt 1] 0]; # cadr
	clipboard append  $txt
    }

    proc paste {w} {
	set txt [selection get -selection CLIPBOARD]
	set tag [lindex $txt 0]
	if [is_not_legal $w.t $tag] {
		tk_dialog  .dialog  "error" "You can't insert this text here" "" 0 "ok"	    
	} else {
	    fromtext $w $txt 0
	}
    }

    proc winpastetext {win} {
	set w [base_window $win]
	pastetext $w
	$win see insert
	upper_tag $win
    }

    proc pastetext {w} {
	global env
	set txt [selection get -selection $env(DEFAULTSELECTION)]
	if [is_not_legal $w.t PCDATA] {
		tk_dialog  .dialog  "error" "You can't insert this text here" "" 0 "ok"	    
	} else {
	    textinsert $w.t $txt
	}
    }

    # As now, the debut and fin marks are only one character long, there is 
    # no need for the following (somehow bugged) procedure

    proc tagchoice {win x y} {
	variable LEGALTAGS
	variable THE_TAGS
	variable VARS
	set m .dicopop
	catch {destroy $m}
 	menu $m
	regsub {\.t$} $win {} w
	set tag [tagname [upper_tag $win]]
	if { $tag == "" } {
	    set w [base_window $win]
	    set tag $VARS($w,roottag)
	}
	
 	foreach subtag $LEGALTAGS($tag) {
	    if { $subtag != "PCDATA" } {
		if { $THE_TAGS($subtag,load) == "" } {
		    $m add command -label $THE_TAGS($subtag,description)\
			-command [list dico::insert_parenthesis $w $subtag ]
		} else {
		    set v dico::
		    append v $THE_TAGS($subtag,load)
		    $m add command -label $THE_TAGS($subtag,description)\
			-command [list $v  $w ]
		}
	    }
	}
	catch {::tk_popup $m $x $y }
    }
    
    proc textleft {win} {
	$win mark set insert "insert  - 1 chars"
	$win see insert
	dico::upper_tag $win
    }

    proc textright {win} {
	$win mark set insert "insert  + 1 chars"
	$win see insert
	dico::upper_tag $win
    }

    proc textup {win} {
	$win mark set insert "insert  - 1 lines"
	$win see insert
	dico::upper_tag $win
    }

    proc textdown {win} {
	$win mark set insert "insert  + 1 lines"
	$win see insert
	dico::upper_tag $win
    }


    proc see_ref {win} {
	variable SUBWIN
	set t [upper_tag $win]
	if [string match REF* $t] {
	    set lims [$win tag prevrange $t insert]
	    set txt [$win get [lindex $lims 0] [lindex $lims 1]]
	    if [regexp {\[(.*)@(.*)\]} $txt dum text coords] {
		seshw::montretexte $text $coords
	    }
	}
    }

    proc set_blue {win} {
	catch {
	    set t [upper_tag $win current]
	    set lims [$win tag prevrange $t current]
	    $win tag add blue [lindex $lims 0] [lindex $lims 1]
	}
    }

    proc unset_blue {win} {
	catch {$win tag remove blue 0.0 end}
    }
    
    ############################################################
    # General utilities functions. Consider lists as sets.

    # dico::set_less --
    #          returns the difference of two sets
    # Arguments:
    #   s1, s2 : lists
    #
    # Results:
    #   s1 - s2, considering s1 and s2 as sets


    proc set_less {s1 s2} {
	set s1 [lsort $s1]
	set s2 [lsort $s2]
	set result {}
	foreach i $s1 {
	    # advance in s2 until (car s2) >= i
	    set cmp [string compare $i [lindex $s2 0]]
	    while {  $cmp > 0 && $s2 != "" } {
		set s2 [lrange $s2 1 end]
		set cmp [string compare $i [lindex $s2 0]]
	    }
	    
	    if { $cmp != 0 || $s2 == ""} {
		lappend result $i
	    } else {
		set s2 [lrange $s2 1 end]
	    }
	}
	return $result
    }
    
    proc norm_hiero {hi} {
	catch {sesh .tmp_s} err
	.tmp_s delete 0.0 end
	.tmp_s insert insert $hi
	set fin [.tmp_s index insert]
	.tmp_s mark set insert 0.1
	set curr [.tmp_s index insert]
	set signes [.tmp_s liste_signes 0.0 end]
	set result {}
	regsub -all -- {&} $signes " " signes
	# Not split. Split is fussy about successive spaces
	#set signes [split $signes {& }]
	return [signes::normaliser_texte $signes 1]
    }

    proc loaddico {{mode pub}} {
	set fic [tk_getOpenFile  -filetypes {{{dico file} {.dic}} {Any *}}]
	if { "$fic" != "" } {
	    set g [dicogauge]
	    load_file $fic $mode
	    destroy $g
	}
    }

    variable gauge

    proc dicogauge {} {
	variable gauge
	set w .gauge
	toplevel $w
	wm title $w [legendes::trans "Dictionnary loading..."]
	label $w.l -width 40 -textvariable dico::gauge
	pack $w.l
	tkwait visibility $w
	return $w
    }

    proc load_file {fic {mode pub}} {
	variable gauge
	set old_mode [seshbase::get_mode]
	seshbase::set_mode $mode

	set f [open $fic]
	set l [read $f]
	close $f

	seshbase::openb dico
	seshbase::openb idx


	foreach {id txt} $l {
	    set gauge $id
	    update idletasks 
	    if { [string compare $id "MAX"] == 0} continue

	    # Two case : 
	    # a) if mode is priv, we are integrating data
	    #    in our own base. If the data belongs to us,
	    #    we need to change the MAX index value
	    # b) if mode is pub, everything is fine.
	    
	    if {[string compare $mode priv] == 0} {
		seshbase::update_max $id
	    }

	    save_orthography $txt
	    set idx [list_tags $txt]
	    seshbase::writeb dico $id $txt
	    set newtags [list_tags $txt]
	    foreach i $newtags {
		set cur [seshbase::read1 idx $i]
		if {[lsearch -exact $cur $id] == -1} { 
		    lappend cur $id
		    seshbase::writeb idx $i $cur
		}
	    }
	}
	seshbase::closeb idx
	seshbase::closeb dico
	seshbase::set_mode $old_mode
    }
}
