# seshTable.tcl
#
#	Deals with tables in the Sesh database.
#	allows saving, loading, navigating in these tables.

#  2001 Serge Rosmorduc
# rosmord@iut.univ-paris8.fr
# quipe langue et littrature gyptienne, EPHE IVme section
#
# 
# voyez le fichier ../LICENCE.txt pour plus d'informations
# See the ../LICENCEE.txt file for more information
#
# $Id$
#

# $Log$
#

# DATA STRUCTURE SUMMARY :
#
# In the database :
#
# any key has a prefix which indicates its kind :
#
# ADM TABLE <TABLENAME> => <TABLE INFO>
# informations about a table
# (more below)
# 
# ADM MAXIDX <TABLENAME> => for tables with an automatic index.
#
# ENTRY TABLE <IDX> => Entry in a table
# IDX <TABLE> <IDX> => indexed pointers in a table, for secondary indexes.
# 
#
# Signed indexes :
# 
# "Signed" means here that someone's ID is used to sign an index entry,
# it doesn't mean +/- like in C.
#
# Indexes are either notSigned or signed.
# notSigned indexes should be reserved for internal values which won't be exported.
# anything exported should be signed.
#
# composite signature : a signature contains two things :
# the user ID, and a machine ID. The user can freely chose the machine ID.
# the idea is that someone may install tksesh on more than one machine. the machine
# ID will allow him to have a coherent database. 
# 
# 
# FUNCTION SUMMARY --
# 
#
# 

# getNewID TAG : returns the current maximal value for a tag.

# get TAG ID : get an element's value

# update TAG ID VALUE

# make TAG VALUE (returns the ID for the created element)

# getTAGIterator TAG [FIRSTID] : get an iterator on a particular tag.
# with optionnal FIRSTID, sends an iterator pointing to the first element with this value.
# if no such element is available, the iterator will end immediatly.


# only after the COMMIT 
#package require stooop 4                                  ;# load stooop package
#namespace import stooop::*                ;# and import class, new, ... commands

package provide seshDB 1.0
package require Dbinter

# A database representation.
# Replaces the original seshbase.tcl file

class SeshDB {
    
    # PUBLIC SeshDB --
    #	Create a database object. A database is contained in 
    #	files, stored in a number of directories.
    #	This allows both personnal and general databases.
    #	For instance, a database may be stored in two files
    #	called "dico", contained in directories /home/foo and /usr/local/db
    #	Reading is done in all files, *in order*, but writing is 
    #	done in only *ONE* file.
    #  
    # Parameters:
    # 	fileName : name of the file containing the base
    #	dirNames : list of directories containing the base
    #	blocksize : basic size of a database entry.
    # Options:
    # Results:
    # Side Effects:
    # Called by:

    proc SeshDB {this fileName dirNames {blocksize 128}} {
	set ($this,fname) $fileName
	set ($this,dirNames) $dirNames
	set ($this,writeMode) 0
	set ($this,isOpened) 0
	set ($this,blocksize) $blocksize
	# Other fields 
	#	($this,fd,$i) : file descriptor for databases
	#	($this,nbrOpened) : number of files in the database (i.e. length of dirNames)
    }

    proc ~SeshDB {this} {
	if [$($this,isOpened)] {
	    close $this
	}
    }

    proc open {this} {
	
	if {$($this,isOpened)} {
	    error "database already opened !!"
	}

	set i 0
	foreach dir $($this,dirNames) {
	    # Open the file for writing if it's the file
	    # designated by writeMode, and else for reading.
	    # Note that writeMode = -1 means readonly
	    
	    set fic [file join $dir $($this,fname)]
	    if {$i == $($this,writeMode)} {
		set ($this,fd,$i) [open_db $fic $($this,blocksize) w]
	    } else {
		set ($this,fd,$i) [open_db $fic $($this,blocksize) r]
	    }
	    incr i
	}
	set ($this,nbrOpened) $i	
	set ($this,isOpened) 1
    }

    proc close {this} {
	for {set i 0} {$i < $($this,nbrOpened)} {incr i} {
	    close_db $($this,fd,$i)
	}
	set ($this,isOpened) 0
    }
    
    proc setWritingMode {this m} {
	set ($this,writeMode) $m
    }

    proc getWritingMode {this} {
	return $($this,writeMode)
    }

    proc writeData {this key data} {
	put_db $($this,fd,[getWritingMode $this]) $key $data
    }

    # Low level function for getting data from the base.
    # returns one item for key "key", first looking in file 0, then
    # in file 1, etc...
    proc getData {this key} {
	set res ""
	for {set i 0} {$i < $($this,nbrOpened)} {incr i} {
	    set res [get_db $($this,fd,$i) $key]
	    if {[string length $res] != 0} {
		return $res
	    }
	}
    }

    # Returns a DBiterator
    
    proc iterator {this prefix {first ""}} {
	return [new DBIterator $this $prefix $first]
    }

    # PRIVATE getFilePath --

    proc getFilePath {this i} {
	return [file join [lindex $($this,dirNames) $i] $($this,fname)]
    }

    proc getNbrFiles {this} {
	return $($this,nbrOpened)
    }

    proc getFD {this i} {
	return $($this,fd,$i)
    }
}


# Private class : a buffer. Used for DBIterator

class DBIBuffer {
    proc DBIBuffer {this db num start} {
	set ($this,db) $db
	set ($this,num) $num
	set ($this,pos) -1
	set ($this,start) $start
	# Number of items to preload
	set ($this,buffersize) 100
	set ($this,orientation) 1
	preload $this $start
    }

    proc ~DBIBuffer {this} {
    }

    proc next {this} {
	set ($this,orientation) 1
	move $this
    }

    proc previous {this} {
	set ($this,orientation) -1
	move $this
    }

    proc getVal {this} {
	#puts "pos : $($this,pos). buffer : $($this,buffer)"
	return [lindex $($this,buffer) $($this,pos)]
    }

    proc hasVal {this} {
	set res [expr {$($this,pos) >= 0 && $($this,pos) < [llength $($this,buffer)]}]
	return $res
    }
    
    # PRIVATE move
    proc move {this} {
	set oldPos $($this,pos)
	incr ($this,pos) $($this,orientation)
	# If the data seeked is not in the buffer :
	if {$($this,pos) < 0 || $($this,pos) >= [llength $($this,buffer)]} {
	    preload $this [lindex [lindex $($this,buffer) $oldPos] 0]
	    incr ($this,pos) $($this,orientation)
	}
	# If it's still not in the buffer, it means we have reached the end.
    }
    
    proc preload {this key} {
	# puts "preloading with $key"
	# get an index :
	set idx [get_db_idx [getFD $this] $key]
	# puts "valeur de dpart [getfrom_db_idx $idx]"
	set toGet $($this,buffersize)
	
	# if traversal is backward-oriented, we first go back :
	if {$($this,orientation) == -1} {
	    # We only go back $toGet - 1 steps, as we want to
	    # keep the current element in range.
	    for {set i 1} {$i < $toGet} {incr i} {
		if {! [incr_db_idx $idx -1] } {
		    incr i -1
		    incr_db_idx $idx 1
		    #puts "getting "
		    #puts "[getfrom_db_idx $idx]"
		    break
		}
	    }
	    set toGet $i
	}
	
	# Now, we try to get $toGet values :

	set l [list]
	for {set i 0} {$i < $toGet} {incr i} {
	    lappend l [getfrom_db_idx $idx]
	    set hasNext [incr_db_idx $idx 1]
	    if {! $hasNext} {
		break
	    }
	}

	set ($this,buffer) $l

	if {$($this,orientation) == -1} {
	    set ($this,pos) [expr [llength $l] - 1]
	} else {
	    set ($this,pos) 0
	}
	free_db_idx $idx
	# SeshDB::close $($this,db)
    }

    proc getFD {this} {
	return [SeshDB::getFD $($this,db) $($this,num)]
    }
}

# An iterator on the database.

class DBIterator {
    
    #	prefix : a prefix all keys looked at should have. prefix is a list
    #	first : the first value to look at.
    
    
    proc DBIterator {this db prefix {first ""}} {
	set ($this,db) $db
	set ($this,prefix) $prefix
	# 
	set ($this,orientation) 1
	
	# create all buffers
	for {set i 0} {$i < [getNbrBuffers $this]} {incr i} {
	    set ($this,buffer,$i) [new DBIBuffer $db $i "$prefix $first"]
	}
	set ($this,hasVal) 0
	set ($this,val) ""
	next $this
    }

    proc ~DBIterator {this} {
	for {set i 0} {$i < [getNbrBuffers $this]} {incr i} {
	    delete $($this,buffer,$i)
	}
    }

    proc next {this} {
	# If orientation changes, advance all buffers
	if {$($this,orientation) == -1} {
	    for {set i 0} {$i < [getNbrBuffers $this]} {incr i} {
		DBIBuffer::next $($this,buffer,$i)
	    }
	}
	
	set ($this,orientation) 1
	
	# The identifier for the best buffer
	set best -1
	set ($this,hasVal) 0

	# Look for the smaller available value in buffers
	for {set i 0} {$i < [getNbrBuffers $this]} {incr i} {
	    if {[DBIBuffer::hasVal $($this,buffer,$i)]} {
		set v [DBIBuffer::getVal $($this,buffer,$i)]
		if {[info exists min]} {
		    if {[string compare $min $v] > 0} {
			set min $v
			set best $($this,buffer,$i)
		    }
		} else {
		    set min $v
		    set best $($this,buffer,$i)
		}
	    }
	}

	# advance this buffer
	if {$best != -1} {
	    DBIBuffer::next $best
	    set ($this,hasVal) 1
	    set ($this,val) $min
	}
    }

    proc previous {this} {
	# If orientation changes, make all buffers go back.
	# If orientation changes, advance all buffers
	if {$($this,orientation) == 1} {
	    for {set i 0} {$i < [getNbrBuffers $this]} {incr i} {
		DBIBuffer::previous $($this,buffer,$i)
	    }
	}

	set ($this,orientation) -1
	
	# The identifier for the best buffer
	set best -1
	set ($this,hasVal) 0
	
	# Look for the greatest value available in buffers.

	for {set i 0} {$i < [getNbrBuffers $this]} {incr i} {
	    if {[DBIBuffer::hasVal $($this,buffer,$i)]} {
		set v [DBIBuffer::getVal $($this,buffer,$i)]
		if {[info exists max]} {
		    if {[string compare $max $v] < 0} {
			set max $v
			set best $($this,buffer,$i)
		    }
		} else {
		    set max $v
		    set best $($this,buffer,$i)
		}
	    }
	}

	# make this buffer move backward
	if {$best != -1} {
	    DBIBuffer::previous $best
	    set ($this,hasVal) 1
	    set ($this,val) $max
	}
    }
    
    proc hasVal {this} {
	return $($this,hasVal)
    }

    proc getVal {this} {
	return $($this,val)
    }
    
    # PRIVATE
    proc getNbrBuffers {this} {
	return [SeshDB::getNbrFiles $($this,db)]
    }
}

class SeshTable {
    
    # Open table in database
    proc SeshTable {this database name} {
	set ($this,db) $database
	set ($this,name) $name
    }

    proc ~SeshTable {this} {}

    proc create {this fields indexes} {
	
    }
    
    proc open {this} {
    }

    proc close {this} {
    }

    proc insert {this val} {
    }

    proc remove {this oid} {
    }

    proc getVals {this oid} {
    }

    proc getVal {this oid} {
    }

    proc getSimpleCursor {this cond} {
	return [new SimpleCursor $this $cond]
    }

    proc getPlainCursor {this} {
	return [new PlainCursor $this]
    }
}

# Basically, a cursor is an iterator.

class Cursor {
    
    proc Cursor {this tbl} {
	set ($this,table) $tbl
    }
    
    proc ~Cursor {this} {}

    virtual proc hasVal {this}

    # gets the cursor's current value.
    virtual proc getVal {this}

    # Advances to next item
    virtual proc next {this}

    # Advances to previous item :
    virtual proc prev {this}
    
    proc getTable {this} {return $($this,table)}
}

class SimpleCursor {

    # Private constructor
    # @param tbl : the table
    # @param cond : the condition put on the cursor.
    # 		    syntax : attr op val.
    #	            where op is one of -eq, -ge, -le, -lt, -gt
    
    proc SimpleCursor {this tbl cond} Cursor {$tbl} {
	
    }

    proc ~SimpleCursor {this} {
    }
}

class PlainCursor {
    
    proc PlainCursor {this tbl} Cursor {$tbl} {
	
    }

    proc ~PlainCursor {this} {
    }
}

# A filtering class :
# only 
class FilterCursor {
    
}

