# seshbase.tcl --
#
#        basic database stuff for sesh.
#        SUBJECT TO CHANGE.
#
# Copyright (c) 1998 Serge Rosmorduc
#

#
# All information about sesh is 
# stored in two databases : pub and priv
# priv is someone's base. pub is a compilation of personnal bases.
# pub should only be built using the function load_file
# export functions will allow information exchange between persons.
# Each "card" will be owned by one person
#
#
# ESSENTIAL FIELDS :
# ID : 
#	the ID is actually made of :
#	a TAG : which identifies the card
#	a userID : which identifies the owner
#	a number.
# SHARED :
#	if a card is shared, any further modification is forbidden.
# 
#
# REMARK :
# This base system should be modified in the future.
# 
#  1998,2000 Serge Rosmorduc,
# quipe EC.art, Laboratoire LRIA, Universit de Paris 8
#
# 15 rue Catulienne
# 93236 St Denis Cedex
# FRANCE
#
# Tout droits rservs.
# 
# voyez le fichier ../LICENCE.txt pour plus d'informations
# See the ../LICENCEE.txt file for more information
#
# $Id: seshbase.tcl,v 1.5 2000/10/31 17:19:28 rosmord Exp $
#
#

#
# $Log: seshbase.tcl,v $
# Revision 1.5  2000/10/31 17:19:28  rosmord
# Mac Adaptation
#
#

# FUNCTION SUMMARY --

# known public functions are : 
#	seshbase::declare base file {size 128}} -- declares a database
#	seshbase::openb base -- opens a base
#	seshbase::closeb base -- closes a base
#	seshbase::writeb base key value -- write data to a base. The base must be opened
#	seshbase::readb base key -- reads data from an opened base
#	seshbase::read1 base key -- reads data from the user part of a database

# 	seshbase::lread -- reads a list of values
# 	seshbase::lwrite -- APPENDS a list to a list of values
#	seshbase::delete -- deletes an element from a base
#


# DATA SUMMARY -- 
# 	Fields of VARS : 
#        VARS(BASE,fname) name of the file for the base
#        VARS(BASE,blockSize) size of a block in the database
#	 VARS(BASE,OWNER,...)
#	 	where BASE = "idx" or "dico" (the base's name)
#	 	and OWNER = pub (global base) or priv (private base)
#	 	: information about bases
#
#
#	 VARS(BASE,OWN,fd) : handle for opened file
#	 VARS(BASE,mode) : opening mode for the base : pub or priv
#	     if "priv", you write in user's base; if "pub", in global base
#	 VARS(BASE,opened) : true if base is already opened.
#	 VARS(mode) : default opening mode for base
#
#	 VARS(maxscan) : scan maximum index
#	 VARS(scan,$id,start) : scan $id first element
#	 VARS(scan,$id,$base,end) : scan $id last element
#	 VARS(scan,$id,$base,tail) : elements read in the base, but not sent
#	 VARS(scan,$id,$base,atend) : true if the scan is finished
#

package provide seshbase 0.0

package require Dbinter
package require seshutils
package require seshuser
package require seshnormalize

namespace eval seshbase {
    variable VARS
    namespace export lread;		# reads a list of values
    namespace export lwrite;		# append a list of values to a list of values
    namespace export read1;		# reads only from one base
    namespace export openb closeb
    namespace export readb;		# reads from the bases (private and public)
    namespace export writeb;		# writes to the default base
    namespace export get_max;		# gets a new index in default base
    namespace export get_mode set_mode;	# choses the default base.
    namespace export isnotmine
    namespace export user_dump;		# dump user database
    
    namespace export open_scan 
    namespace export close_scan
    namespace export scanget
    


    variable USERID
    # mode = 1 (user) or 0 (supervisor)

    # default value for mode : private.


    # seshbase::declare --
    # 	declares a base
    # Parameters:
    #	base: name of the base
    #   file: filename associated with the base
    # Options:
    # Results:
    # Side Effects:
    # 	the said base is declared. Nothing is created yet.
    # Use:
    #	PUBLIC
    # Called by:
    
    proc declare {base file {size 128}} {
	variable VARS
	set VARS($base,fname) $file
	set VARS($base,blockSize) $size
	set VARS($base,opened) 0
    }

    # seshbase::getFileName --
    # 	get the path to the file associated with a base
    #   either in public or private mode
    # Parameters:
    #	base: name of the base
    #   own: owner of the base (pub or priv)
    # Options:
    # Results:
    # 	the path to the file
    # Side Effects:
    # Use:
    #	PRIVATE
    # Called by:
    

    proc getFileName {base owner} {
	variable VARS
	switch -exact -- $owner {
	    pub {
		return [file join $::env(SESHLIB) $VARS($base,fname)]
	    }
	    priv {
		return [file join [seshuser::getvar PATH] $VARS($base,fname)]
	    }
	}
    }
    
    # seshbase::openb --
    # 	opens a base
    # Parameters:
    #	base : either idx for index or dico for the dictionnary
    # Options:
    # Results:
    # Side Effects:
    # 	the said base is opened and can be written to or read,
    #   plus, it's created if it doesn't already exist
    # Use:
    #	PUBLIC
    # Called by:

    proc openb  {base} {
	variable VARS
	set VARS($base,pub,fd) -1
	set VARS($base,priv,fd) -1
	set VARS($base,mode) $VARS(mode)
	set mode $VARS(mode)

	set f1 [getFileName $base pub]
	set f2 [getFileName $base priv]
	set taille $VARS($base,blockSize)

	if {[info exists VARS($base,opened)] && $VARS($base,opened)} {
	    error "database already opened !!"
	}

	set VARS($base,opened) 1

	if {[string compare $mode "priv"] == 0} {
	    if [catch {set VARS($base,pub,fd) [open_db $f1 $taille r]}] {
		set VARS($base,pub,fd) -1
	    }
	    set VARS($base,priv,fd) [open_db $f2 $taille w]
	} else {
	    set VARS($base,pub,fd) [open_db $f1 $taille w]
	    if [catch {set VARS($base,priv,fd) [open_db $f2 $taille r]}] {
		set VARS($base,priv,fd) -1
	    }
	}
    }


    # seshbase::closeb --
    # 	closes a base
    # Parameters:
    #	base : either idx for index or dico for the dictionnary
    # Options:
    # Results:
    # Side Effects:
    # 	the said base is closed
    # Use:
    #	PUBLIC
    # Called by:
    
    proc closeb {base} {
	variable VARS
	if {$VARS($base,pub,fd) != -1 } {
	    close_db $VARS($base,pub,fd)
	}
	if {$VARS($base,priv,fd) != -1 } {
	    close_db $VARS($base,priv,fd)
	}
	set VARS($base,opened) 0
    }

    # seshbase::writeb --
    # 	write data to a base. The base must be opened
    # Parameters:
    #	base : either idx for index or dico for the dictionnary
    #	key : the key for the data
    #	value : the value stored
    # Options:
    # Results:
    # Side Effects:
    # 	the data is written in the correct base (pub or priv)
    # Use:
    #	PUBLIC
    # Called by:

    proc writeb {base key value} {
	variable VARS
	set cible $VARS($base,mode)
	put_db $VARS($base,$cible,fd) $key $value
    }
    
    # seshbase::readb --
    # 	reads data from an opened base
    # Parameters:
    #	base : either idx for index or dico for the dictionnary
    #	key : the key for the data
    # Options:
    # Results:
    # 	the data associated to key. If data exists in private database,
    # 	only this one is returned
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:

    proc readb {base key} {
	variable VARS
	set res ""
	if { $VARS($base,priv,fd) != -1 } {
	    set res [get_db $VARS($base,priv,fd) $key]
	}
	if {[string length $res] != 0} {
	    return $res
	} else {
	    if {$VARS($base,pub,fd) != -1} {
		set res [get_db $VARS($base,pub,fd) $key]
	    }
	    return $res
	}
    }

    # seshbase::delete --
    # 	delete data from a base. The base must be opened
    # Parameters:
    #	base : either idx for index or dico for the dictionnary
    #	key : the key for the data
    # Options:
    # Results:
    # Side Effects:
    # 	the data is deleted from the correct base (pub or priv)
    # Use:
    #	PUBLIC
    # Called by:
    
    proc delete {base key} {
	variable VARS
	set cible $VARS($base,mode)
	delete_db $VARS($base,$cible,fd) $key
    }

    # seshbase::lread --
    # 	reads a list  from an opened base
    # Parameters:
    #	base : either idx for index or dico for the dictionnary
    #	key : the key for the data
    # Options:
    # Results:
    # 	the list of data associated to key. if data exists in both
    # 	bases, the whole data is returned
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:

    proc lread {base key} {
	variable VARS
	set res1 "" 
	set res2 ""
	if { $VARS($base,priv,fd) != -1 } {
	    set res1 [get_db $VARS($base,priv,fd) $key]
	}
	if { $VARS($base,pub,fd) != -1 } {
	    set res2 [get_db $VARS($base,pub,fd) $key]
	}
	return [concat $res1 $res2]
    }



    # seshbase::lwrite --
    # 	writes a list  to an opened base
    # Parameters:
    #	base : either idx for index or dico for the dictionnary
    #	key : the key for the data
    #	values : the values for the data
    # Options:
    # Results:
    # Side Effects:
    # 	the list of values is appended to the database entry for key
    # Use:
    #	PUBLIC
    # Called by:

    proc lwrite {base key values} {
	variable VARS
	set cible $VARS($base,mode)
	set vals [get_db $VARS($base,$cible,fd) $key]
	set vals [concat $vals $values]
	set vals [uniq $vals]
	put_db $VARS($base,$cible,fd) $key $vals
    }


    # seshbase::read1 --
    # 	reads data from the database opened for writing 
    #	(normally the user database).
    # Parameters:
    #	base : either idx for index or dico for the dictionnary
    #	key : the key for the data
    # Options:
    # Results:
    # 	the data associated to key
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:
    
    proc read1 {base key} {
	variable VARS
	set where $VARS($base,mode)
	set res [get_db $VARS($base,$where,fd) $key]
	return $res
    }

    # seshbase::set_mode --
    # 	sets the default access mode for databases
    # Parameters:
    #	mode : either pub or priv
    # Options:
    # Results:
    # Side Effects:
    # 	writing will take place in the public (pub) or
    #   private (priv) database. private is the default,
    #	as modification normally take place only in the user's own
    #	database. Public writing should be only called for administrative tasks.
    # Use:
    #	PUBLIC
    # Called by:

    proc set_mode {mode} {
	variable VARS
	set VARS(mode) $mode
    }

    # seshbase::get_mode --
    # 	gets the current  default access mode for databases
    # Parameters:
    # Options:
    # Results:
    #	the current  default access mode for databases (pub or priv)
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:

    proc get_mode {} {
	variable VARS
	return $VARS(mode)
    }

    # seshbase::get_max --
    #	Get a new index for current user
    # 	dico should have been opened first
    #   THIS FUNCTION SHOULD EITHER MOVE OR GET MORE GENERAL
    # Parameters:
    # Options:
    # Results:
    #	a new numeric index value
    # Side Effects:
    #	the max value stored in the database will be incremented
    #	next call will return an incremented value
    # Use:
    #	PUBLIC
    # Called by:

    proc get_max {} {
	set id [seshuser::getvar ID]
	set MAX  [read1 dico MAX]
	if { $MAX == "" } {
	    set MAX 0
	}
	incr MAX
	seshbase::writeb dico MAX $MAX
	return ${id}_${MAX}
    }


    # seshbase::getMaxID --
    #	Get a new index for the current user and a given tag
    # 	dico should have been opened first
    # Parameters:
    # Options:
    # Results:
    #	the current value of MaxID
    # Side Effects:
    #	None
    # Use:
    #	PUBLIC
    # Called by:

    proc getMaxID {TAG} {
	set id [seshuser::getvar ID]
	set MAX  [read1 dico MAX::$TAG]
	if { $MAX == "" } {
	    set MAX 0
	}
	incr MAX
	# seshbase::writeb dico MAX::$TAG $MAX
	return $MAX
    }
    
    proc requestID {TAG ID} {
	set m [getMaxID $TAG]
	if {$m <= $ID} {
	    seshbase::writeb dico MAX::$TAG $ID
	}
    }
 
    # seshbase::isnotmine --
    #	tests if an id record for a database entry belongs to the current user
    # Parameters:
    #	the id record entry
    # Options:
    # Results:
    #	true if the id record doesn't belong to the current user,
    #	else false
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:

    proc isnotmine {idrecord} {
	set id [seshuser::getvar ID]
	if [regexp "^${id}_" $idrecord] {
	    return 0
	} else {
	    return 1
	}
    }


    # seshbase::update_max --
    #	knowing an index for an entry, makes sure
    #	the user's index is currently up-to-date
    #	The purpose of this function is to ease the
    #	importation of old version of the user's database
    #
    #	CURRENTLY LINKED TO THE dico DATABASE
    #
    # Parameters:
    #	idx: index in the database
    # Options:
    # Results:
    # Side Effects:
    #	the max value stored in the database will be updated
    #	if needed
    # Use:
    #	PUBLIC
    # Called by:

    
    proc update_max {idx} {
	set id [seshuser::getvar ID]
	if [regexp "^${id}_(\[0-9\]*)\$" $idx tmp nbr] {
	    set MAX [read1 dico MAX]
	    if { $MAX == "" } {
		set MAX 0
	    }
	    if {$nbr > $MAX} {
		seshbase::writeb dico MAX $nbr
	    }
	}
    }

    
    # seshbase::user_dump --
    #	dump a database in "readable" (text) format
    # Parameters:
    #	base: the database name
    #	owner: pub or priv.
    # Options:
    # Results:
    #	the database in textual format
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:

    proc user_dump {base {owner priv}} {
	set f [getFileName $base $owner]
	set db [open_db $f]
	set txt [dump_db $db]
	close_db $db
	return $txt
    }

    ######################################################################
    # scan of a base  :
    # 
    # open_scan TAG <start> --
    #   sends back an identifier
    # close_scan <id> --
    #   end of the scan
    #
    # scanget <id> <num> --
    #   get (at most) <num> elements from the  base,
    #   starting with the current starting point.
    #   The next starting point is the last element of this list ????
    #
    # isscanend <id> --
    #   are there other elements in the scan ?
    #
    # variables used : 
    #
    # the first and last elements of a scan are used
    #
    # VARS(maxscan) : scan maximum index
    # VARS(scan,$id,start) : scan $id first element
    # VARS(scan,$id,$base,end) : scan $id last element
    # VARS(scan,$id,$base,tail) : elements read in the base, but not sent
    
    set VARS(maxscan) 0

    # seshbase::open_scan --
    #	open a new scan in the database.
    #	a scan is an abstract iterator that can be used
    #	to list entries in the database in order.
    #
    #	Entries are stored and indexed by TAG values
    #
    #
    # Parameters:
    #	TAG: the tag which will be used
    #	start: the first value of the TAG.
    #	base: the database name
    # Options:
    # Results:
    #	an value which represent this particular scan
    # Side Effects:
    #	a scan is initiated. The internal C representation makes scans a 
    #	short-lived structure (not supporting data modification in the middle), but this 
    #	tcl wrapper simulate robust scans and can be used without restriction.
    # Use:
    #	PUBLIC
    # Called by:


    proc open_scan {TAG start {base idx}} {
	variable VARS
	set id  [incr VARS(maxscan)]
	set VARS(scan,$id,start) $start
	set VARS(scan,$id,pub,end) $start
	set VARS(scan,$id,priv,end) $TAG
	set VARS(scan,$id,tag) $TAG
	set VARS(scan,$id,pub,atend) 0
	set VARS(scan,$id,priv,atend) 0
	set VARS(scan,$id,pub,tail) {}
	set VARS(scan,$id,priv,tail) {}
	set VARS(scan,$id,base) $base
	return $id
    }

    # seshbase::close_scan --
    #	closes a scan in the database.
    # Parameters:
    #	id: the scan id which was returned by open_scan
    # Options:
    # Results:
    # Side Effects:
    #	the scan is closed
    # Use:
    #	PUBLIC
    # Called by:

    proc close_scan {id} {
	variable VARS
	unset VARS(scan,$id,start)
	unset VARS(scan,$id,pub,end)
	unset VARS(scan,$id,priv,end)
	unset VARS(scan,$id,tag)
	unset VARS(scan,$id,pub,atend)
	unset VARS(scan,$id,priv,atend)
	unset VARS(scan,$id,pub,tail)
	unset VARS(scan,$id,priv,tail)
	unset VARS(scan,$id,base)
	
	#unset VARS(scan,$id,start) 
	#unset VARS(scan,$id,priv,end)
	#unset VARS(scan,$id,pub,end)
    }

    # seshbase::scanaux --
    # 	auxiliary function used by scanget.
    #	gets at most num elements from one part 
    #	(public or private) of the base.
    #
    # 	btdb is rather stupid when it comes to non existing files.
    # 	we are careful to test every option here. Anything opened
    # 	is closed.
    # Parameters:
    # 	id: scan id
    # 	owner: pub or priv
    # 	num: number of elements to get
    # Options:
    # Results:
    #	the elements
    # Side Effects:
    # Use:
    #	PRIVATE
    # Called by:
    #	scanget

    #
    # NOTE : We should get sure there is no duplicate entry.
    # The problem is that db sends back the first value >= key
    # thus, if the key is already part of the list,
    # we retrieve it.
    #
    # Now, the idea is that we always drop the first element.
    # if this is not the first call to scanaux.
    
    proc scanaux {id owner num} {
	variable VARS

	set res $VARS(scan,$id,$owner,tail)

	if {[llength $res] >= $num} {
	    return $res
	}
	
	set base $VARS(scan,$id,base)
	openb $base
	set db $VARS($base,$owner,fd)

	if {$db == -1} {
	    set res {}
	} else {
	    set good [expr ! $VARS(scan,$id,$owner,atend)]
	    # puts "I am Looking for $VARS(scan,$id,$owner,end)"
	    if [catch {set i [get_db_idx $db $VARS(scan,$id,$owner,end)]}] {
		set res ""
	    } else {
		# puts "HERE $i $num $good"
		for {set j 0} {$j < $num && $good} {incr j} {
		    set last [getfrom_db_idx -key $i]
		    # puts ":::$last"
		    set tag_is_bad [string compare \
			    [lindex $last 0] \
			    $VARS(scan,$id,tag)]
		    
		    if $tag_is_bad {
			set good 0
			set VARS(scan,$id,$owner,atend) 1
		    } else {
			lappend res $last
			set VARS(scan,$id,$owner,end) $last
			set good [incr_db_idx $i 1]
		    }
		}
		if {! $good} {
		    set VARS(scan,$id,$owner,atend) 1
		}
		free_db_idx $i
	    }
	}
	closeb $base
	return $res
    }

    # seshbase::scanget --
    #	gets at most num elements from the base
    # Parameters:
    #	id: the scan id which was returned by open_scan
    #	num: number of elements to get (if possible)
    # Options:
    # Results:
    # Side Effects:
    #	the scan advances
    # Use:
    #	PUBLIC
    # Called by:

    proc scanget {id num} {
	variable VARS
	set n1 [expr $num + 1]
	set l1 [scanaux $id pub $n1]
	set l2 [scanaux $id priv $n1]

	# now, if possible, l1 and l2 contain 
	# enough elements.
	set i 0
	set res {}
	#puts "L1>> $l1"
	#puts "L2>> $l2"

	while {$i < $num && [not_empty l1 l2]} {
	    get_best res l1 l2
	    incr i
	}
	#puts "L1post>> $l1"
	#puts "L2post>> $l2"
	set VARS(scan,$id,pub,tail) $l1
	set VARS(scan,$id,priv,tail) $l2

	return $res
    }

    # seshbase::not_empty
    #	auxilliary function to test if there are elements left in a scan
    # Parameters:
    #	l1, l2: name of lists variables
    # Options:
    # Results:
    #	true if one of l1, l2 is not empty
    # Side Effects:
    # Use:
    #	PRIVATE
    # Called by:
    #	scanget

    proc not_empty {l1 l2} {
	upvar $l1 la
	upvar $l2 lb
	return [expr [llength $la] != 0 || [llength $lb] != 0]
    }

    # seshbase::not_empty
    #	auxilliary function to get the best element of list l1 and l2
    #   in the result list res
    #   the best element of l1 and l2 is the least element in alphabetic
    #	order. The function takes care of special cases (l1 and/or l2 empty)
    # Parameters:
    #	res: the result list
    #	l1, l2: name of lists variables
    # Options:
    # Results:
    # Side Effects:
    #	l1, l2, and res are modified.
    # Use:
    #	PRIVATE
    # Called by:
    #	scanget

    proc get_best {res l1 l2} {
	upvar $l1 la
	upvar $l2 lb
	upvar $res liste
	set lgtha [llength $la]
	set lgthb [llength $lb]
	
	if {$lgtha != 0 && $lgthb != 0} {
	    set a [lindex $la 0]
	    set b [lindex $lb 0]
	    if {[string compare $a $b] < 0} {
		# we take a
		lappend liste $a
		set la [lrange $la 1 end]
	    } else {
		# we take b
		lappend liste $b
		set lb [lrange $lb 1 end]
	    }
	} else {
	    if {$lgtha != 0} {
		lappend liste [lindex $la 0]
		set la [lrange $la 1 end]
	    }
	    if {$lgthb != 0} {
		lappend liste [lindex $lb 0]
		set lb [lrange $lb 1 end]
	    }
	}
    }

    
    proc isscanend {id} {
	error "not written yet. Maybe never"
    }
}


catch {
    seshbase::set_mode priv
    seshbase::declare idx newdic_IDX 16
    seshbase::declare dico newdic
    seshbase::declare texts texts
}

