# signes.tcl --
#
#       Common data about hieroglyphic signs
# 
#
# Remark : this file correspond to the time when signs where hard-coded
# now that they are defined in signedef, we should
# probably have a slightly different system.
#
# 
######################################################################
#
# Notes :
#
# Sign variant system :
# -------------------
#
# We shall use here a system with two levels of variants :
#
# a) sure variants : like Y1v for Y1, D40 for A24, Y2 for Y1 &.
#    These are variants which might mean something (for example, 
#    indicate a  hieratic original or date the text),
#    but don't lead to confusion. We don't use these variants 
#    for indexing the texts, however, because one might want 
#    to distinguish between them.
#
# b) synsets : the idea  is taken from a lexicology program called wordnet.
# a synset is a set of "synonymous signs". The synonymy relation
# is NOT transitive, however. A sign might belong to different synsets.
# 
# For example, there are a number of synset which contain D36 (the arm).
# One contains D46 (the hand), and another contains D40, but this does not make
# D46 and D40 synonymous signs. Typically, this relation can not be used in a
# full automatic way. One should be allowed to chose either a sign or a synset.
#
# Word looking system :
# ---------------------
#
# This system takes as its argument a list of signs. It can then either
#
# a) build a regular expression usable to look for the word.
# b) send back a list of equivalent writings
#
# The argument can contain :
#
# - signs : exact sign matching (gardiner code)
# - list of signs : matches one signe in the list
# - a "?" can precede the sign or list, in which case it's optionnal.
#
# e.g. : F40 ? G1 ? {G43 Z7} Y1
# will correspond to "Aw-Y1" or to "Aw-W-Y1", etc.
#
# implementation note :
#   The regular expression should be able to span lines.
#   between signs, we can have either a space, or "\n[^\t]*\t"
#   (the line number)

#package require seshutils
package provide signes 0.0




namespace eval signes {
    variable signesF;			# fonts of signs
    variable signesN;			# number of sign in font
    variable signesP;			# phonetic value of signs
    variable codes;			# codes of defined signs
    variable variants;			# table of base for for signs which are variants
    variable synset;
    variable normcode;			# table of normal values for signs.
    
    # get the list of possible signs for a given code
    namespace export code_to_signes
    # get possible phonetic values for a given sign
    namespace export values_for_sign 
    # normalize a list of sign.
    namespace export normaliser_texte
    # normalize one sign.
    namespace export normalize_sign
    namespace export get_varset
    namespace export define_synset
    namespace export get_all_syns
    namespace export list_to_regexp
    namespace export expand_list
    # register a sign code
    namespace export register_sign
    # get the name of bitmap files for a sign
    namespace export fontbitmapfile nom_bitmap
    # list of sign in a given list (A, B, C...)
    namespace export signes_in_lettre
    # declares sign code1 as a variant of code2
    namespace export define_variant
    ; # code to font number
    array set signesF {}
    ; # code to char number in font
    array set signesN {}
    ; # phonetic values and codes to base code.
    array set signesP {}
    array set variants {}
    array set normcode {}

    set synset(n) 0;			# max synset number

    catch {
	source [file join  $env(SESHLIB) signearray.tcl]
    }

    # the bitmap file which contains the sign's drawing
    proc fontbitmapfile {signe} {
	variable signesF
	variable signesN
	variable signesP
	return $signesF($signe)/$signesF($signe).$signesN($signe).pbm
    }
    
    # the name of the Tcl bitmap associated with the sign in menus.
    proc nom_bitmap {signe} {
	global bitmapsd
	variable signesN 
	variable signesF
	return $signesF($signe)$signesN($signe)
    }

    # Ordering function for signs
    proc comparesignes {s1 s2} {
	array set rgltr {A 1 B 2 C 3 D 4 E 5 F 6 G 7 H 8 I 9 J 10 K 11 L 12 M 13 N 14 \
			     O 15 P 16 Q 17 R 18 S 19 T 20 U 21 V 22 W 23 X 24 Y 25 Z 26 Aa 27}
	regexp {([A-Z]a?)([0-9]+)} $s1 dummy LTR1 COD1
	regexp {([A-Z]a?)([0-9]+)} $s2 dummy LTR2 COD2
	set ltrcmp [expr $rgltr($LTR1) - $rgltr($LTR2) ]
	if [expr $ltrcmp != 0] {
	    return $ltrcmp
	} else {
	    return [expr $COD1 - $COD2]
	}
    }

    # List of signs in a section (A, B, etc...)
    
    proc signes_in_lettre {lettre} {
	variable signesF
	return [lsort -command comparesignes \
		    [array names signesF [append lettre {[0-9]*}]]]
    }

    proc values_for_sign {sign} {
	variable signesP
	set texte [array names signesP $sign=* ]
	set val ""
	scan $texte {%[^=]=%s} dummy val
	return $val
    }

    proc code_to_signes {code} {
	variable signesF
	variable signesN
	variable signesP
	variable codes

	if [info exists codes($code)] {
	    set liste [list $code] 
	} else {
	    set liste {}
	}

	# rcuprer la liste de tous les signes correspondant
	set listeaux0 [array names signesP *|$code|* ]
	foreach i $listeaux0 {
	    lappend liste $signesP($i)
	}

	# voir si le signe est un code repertori
	if {[info exists signesF($code)]} {
	    # supprimer code de la liste 
	    set indexcode [lsearch -exact $liste $code]
	    if { $indexcode != -1} {
		set liste [lreplace $liste $indexcode $indexcode]
	    }
	    set liste [concat $code $liste]
	}
	return $liste
    }
  
    proc register_sign {code} {
	variable codes
	set codes($code) 1
    }

    proc normaliser_texte {txt {minimal 1}} {
	set result ""
	if $minimal {
	    set norm normalize_sign
	} else {
	    set norm low_level_norm
	}
	foreach i $txt {
	    lappend result [$norm $i]
	}
	return $result
    }

    # declares sign code1 as a variant of code2
    proc define_variant {code1 code2} {
	variable variants
	set variants($code1) $code2
	lappend variants(varset,$code1) $code2
	lappend variants(varset,$code2) $code1
    }

    proc get_varset {code} {
	variable variants
	if [info exists variants(varset,$code)] {
	    set l $variants(varset,$code)
	    lappend l $code
	    return $l
	} else {
	    return $code
	}
    }

    proc define_synset {signs} {
	variable synset
	set l [normaliser_texte $signs 0]
	set n [incr synset(n)]
	set synset($n,signs) $l
	foreach s $l {
	    lappend synset($s,nums) $n
	}
    }

    proc get_all_syns {code} {
	variable synset
	if [info exists synset($code,nums)] {
	    foreach i $synset($code,nums) {
		foreach s $synset($i,signs) {
		    lappend result $s
		}
	    }
	    return $result
	} else {
	    return $code
	}
    }
    
    # get a list of signs and sign group and generate all
    # possibilities
    proc expand_list {l} {
	set result {{}}
	if [llength $l] {
	    set opt 0
	    set h [lindex $l 0]
	    set l [lrange $l 1 end]
	    if { [string compare $h "?"] == 0} {
		set opt 1
		set h [lindex $l 0]
		set l [lrange $l 1 end]
	    }

	    set tails [expand_list $l]
	    
	    if $opt {
		# as the head element is optional, the result
		# is initialized as :
		set result $tails
	    } else {
		set result {}
	    }

	    foreach x $h {
		foreach e $tails {
		    lappend result [concat $x $e]
		}
	    }
	}
	return $result
    }

    # get a list of signs and sign group and create the corresponding regular expression.
    # 
    # The expression is a bit cumbersome. It might be bettered in two ways :
    # 
    # a) non optional lone signs don't need parenthesis at all
    # b) optional lone signs need only one set of parenthesis
    #
    # The  result can be used in various regexps
    #
    # e.g. transform a text into a list of 
    # the lines which contain the signs :
    #
    # set res [list_to_regexp ...]
    # regsub -all "(\[^\t\n\]+\t)\[^\n\]*$res\[^\n\]*|\[^\n\]*\n" $txt {\1}

    proc list_to_regexp {l} {
	#set space "( |(\n|\r|\r\n)\[^\t\]*\t)"
	set space "( | \n\[^\t\]*\t)"
	set r ""
	set opt 0
	foreach e $l {
	    if {[string compare $e "?"] == 0} {
		set opt 1
	    } else {
		append r "("
		append r "("
		append r [lindex $e 0]
		foreach x [lrange $e 1 end] {
		    append r "|"
		    append r $x
		}
		append r ")"
		append r " )"
		if $opt {
		    append r ?
		}
		set opt 0
	    }
	}
	regsub " (\[^ \]*)$" $r {\1} r
	regsub -all " " $r $space r
	return $r
    }

    # is_gardiner is true if code is NOT a phonetic code
    proc is_gardiner {code} {
	return [regexp {^[A-Za-z]+[0-9]+[A-Za-z]*} $code]
    }

    # Low level norm gives, for one sign, its canonical Gardiner form.
    # We should probably use it to generate a precomputed list of such forms
    # As we are too lazy to do it, we shall use some caching :
    
    proc low_level_norm {code} {
	variable signesF
	variable signesN
	variable normcode
	if [info exists normcode($code)] {
	    return $normcode($code)
	}
	set result $code
	switch -exact -- $code {
	    W {set result "Z7"}
	    N {set result "S3"}
	    M {set result "Aa13"}
	    default {
		if {! [is_gardiner $code]} {
		    if [info exists signesF($code)] {
			set x $signesF($code)$signesN($code)
			# x is the character as found in the fonts.
			# Now, we look at candidate signs :
			set found 0
			foreach cand [code_to_signes $code] {
			    if [is_gardiner $cand] {
				if [info exists signesF($cand)] {
				    set y $signesF($cand)$signesN($cand)
				    if {[string compare $x $y] == 0} {
					set found 1
				    }
				}
			    }
			    if $found break
			}
			if $found {
			    set result $cand
			}
		    }
		}
	    }
	}
	set normcode($code) $result
	return $result
    }
    
    proc normalize_sign {code} {
	variable variants
	# a) first, we get Gardiner code.
	# 
	set code [low_level_norm $code]
	if [info exists variants($code)] {
	    return $variants($code)
	} else {
	    return $code
	}
    }
}
