# stext.tcl --
#
#       (extensible) Structured text editor in tcl
#
#	stext defines a simple way to describe structured text
#	syntax and semantics via tcl. a widget built from stext will 
#       be able to edit structured text -- even with complex subwidget, and in a "wysiwyg" way,
#       to display them, save them (in a number of formats), load them, etc.
#       
#
# 
#  2000 Serge Rosmorduc,
# quipe EC.art, Laboratoire LRIA, Universit de Paris 8
#
# 15 rue Catulienne
# 93236 St Denis Cedex
# FRANCE
#
# Tout droits rservs.
# 
# voyez le fichier ../LICENCE.txt pour plus d'informations
# See the ../LICENCEE.txt file for more information
#
# $Id$
#
#

#
# $Log$
#

# FUNCTION SUMMARY --
#
# known public functions are : 
# 	
# DATA SUMMARY -- 
#
# Variables :
# VARS($w,maxtag)	: max used id for a tag
# VARS($w,elide) 	: should hiden data be shown ?
# VARS($w,doctype)	: document type of a window
# VARS($w,$tagid,...) : information about tags (tag id are numbers)
# VARS($w,$num,type)	: type of tag
# VARS($w,$num,parent)	: parent tag
# VARS($w,$num,opt,$o) : option of tag

# VARS($doctype,$tag,opt,$o) : option for a tag description in a document type
#
# Note : tag id 0 is reserved for the whole document
# 
# Tag structure.
#
# tags are used like this :
#
# INTEGER : these tags correspond to a single tag in the text.
# "btag INTEGER" mark the beginning of the INTEGER tag
# "etag INTEGER" mark the end of the INTEGER tag
# 
# Embedded tags and text attribute inheritance :
# It's rather easy. As the only tag that matter for structure purporses are
# begin and and tags, we simply keep all other tags.

package provide stext 0.0	;# This uses Tcl namespaces

package require myopts

namespace eval stext {
    
    # VARS contains data for each stext window.
    # For window .w :
    # VARS(.w,NBR) maximum tag number
    # VARS(.w,ROOT) root tag for this window
    variable VARS

    # stext::create --
    #	creates a new window for displaying stext
    # Parameters:
    #	w	: window name
    #	type	: document type edited in the window.
    #		  must have been created using doctype::create
    # Options:
    # Results:
    #	returns the created window
    # Side Effects:
    #	a nice window is created
    # Use:
    #	PUBLIC
    # Called by:
    
    proc create {w type} {
	variable VARS
	frame $w
	text $w.t
	grid $w.t

	bindtags $w.t [list $w.t HierarchicText all] 	
	
	set VARS($w,elide) 0
	set VARS($w,maxtag) 1
	set VARS($w,doctype) $type
	
	# Tag 0 is document
	set VARS($w,0,type) $type
	toggleElide $w
	clear $w
	insertNewTagContent $w $type insert
	return $w
    }

    # stext::delete --
    #	deletes a stext window
    # Parameters:
    #	w	: window name
    # Options:
    # Results:
    # Side Effects:
    #	w is deleted.
    # Use:
    #	PUBLIC
    # Called by:

    proc delete {w} {
    }

    # stext::clear --
    # Parameters:
    #	w: window name
    # Options:
    # Results:
    # 	a list representation of the window content
    # Side Effects:
    #	None
    # Use:
    #	PRIVATE
    # Called by:
    
    proc clear {w} {
    }

    # stext::toggleElide --
    # Parameters:
    #	w: window name
    # Options:
    # Results:
    # Side Effects:
    #	hidden tag limits are shown or hiden, depending on the
    #	previous state. For debugging purposes
    # Use:
    #	PRIVATE
    # Called by:
    
    proc toggleElide {w} {
	variable VARS
	set VARS($w,elide) [expr 1 - $VARS($w,elide)]
	return
	if [catch {
	    $w.t tag configure btag -elide $VARS($w,elide) -background red
	    $w.t tag configure etag -elide $VARS($w,elide) -background red
	}] {
	    $w.t tag configure btag -background red
	    $w.t tag configure etag -background red
	}
    }

    # stext::textw --
    # Parameters:
    #	w: window name
    # Options:
    # Results:
    #	the text widget associated with the stext one.
    # Side Effects:
    # Use:
    #	PUBLIC ?
    # Called by:
    
    proc textw {w} {
	return $w.t
    }

    # stext::makeList --
    #	build a list representation of a given window content
    # Parameters:
    #	w: window name
    #	start	: index of the beginning of the dumped text
    #	end	: index of the end of the dumped text
    #	all	: does the text correspond to the whole window ???
    # Options:
    # Results:
    # 	a list representation of the window content
    # Side Effects:
    #	None
    # Use:
    #	PRIVATE
    # Called by:
    
    proc makeList {w {start 0.0} {end end} {all 0}} {
	puts hello
	set d [dumptext $w]
	puts "dump : $d"
	set r [makeListAux $w $d]
	puts "result : $r"
	return $r
    }
    
    proc makeListAux {w txt} {
	# I don't know if this is faster than a good
	# old while loop.
	set tstack {}		;# stack of tag names
	set vstack {}		;# stack of tag values
	set current {}		;# current tag value
	foreach {key val} $txt {
	    switch -exact -- $key {
		text {
		    lappend current [list pcdata $val]
		}
		btag {
		    set tstack [list $val $tstack]
		    set vstack [list $current $vstack]
		    set current {}
		}
		etag {
		    # get tag id
		    set tname [lindex $tstack 0]
		    # restore tag stack
		    set tstack [lindex $tstack 1]
		    # copy tag value
		    set myval $current
		    # restore current
		    set current [lindex $vstack 0]
		    # restore tag values stack
		    set vstack [lindex $vstack 1]
		    # TO BE IMPROVED
		    lappend current [list $tname $myval]
		}
		window {
		}
	    }
	}
#  	if {[llength $vstack]} {
#  	    set tmp $current
#  	    set current [lindex $vstack 0]
#  	    lappend current $tmp
#  	}
	# Now, if tstack is not empty, it means we should close the said tags.
	# Well, this SHOULDN'T BE, as each tag has a start tag and an
	# end tag item which is one character long. Now, if we select
	# a text only on tag boundaries, we can't have a tagon or a tagoff for any other tag.
	return $current
    }

    # stext::dumptext --
    #	The dump command of the text widget sends back 
    # 	a number of irrelevant text separations. We 
    # 	deal with these.
    # Parameters:
    #	w: window name
    #	start	: index of the beginning of the dumped text
    #	end	: index of the end of the dumped text
    #	all	: does the text correspond to the whole window ???
    # Options:
    # Results:
    # 	a list representation of the window content.
    #	The list can contain :
    #		text VALUE
    #		btag VALUE
    #		etag VALUE
    # Side Effects:
    #	None
    # Use:
    #	PRIVATE
    # Called by:
     #	makeList
    
    proc dumptext {w {debut 0.0} {fin end}} {
	set txt [$w.t dump -text -tag -window $debut $fin]
	set last {}
	set tmp {}
	set lastWasText 0
	# We build a canonized version of the dump,
	# gluing together the text parts which may be separated because
	# of marks.
	# This is also the place to suppress a number of tags, like
	# sel and the like.
	foreach {k v i} $txt {
	    switch --  $k {
		text {
		    append last $v
		    set lastWasText 1
		}
		tagon -
		tagoff {
		    switch -exact -- $v {
			sel {}
			default {
			    if {[string length $last] != 0} {
				lappend tmp text $last 0
				set last {}
				set lastWasText 0
			    }
			    lappend tmp $k $v $i
   			}
		    }
		}
		window {
		    if {[string length $last] != 0} {
			lappend tmp text $last 0
			set last {}
			set lastWasText 0
		    }
		    lappend tmp $k $v $i
		}
	    }
	}

	# Now, if all tags are added using insert, the last newline
	# should always be there, and untagged. We shouldn't add it.

	if {$lastWasText && [string compare $fin "end"] != 0} {
	    lappend tmp text $last 0
	    set last {}
	}
	

	# Now, suppress btags and etags :
	# NO. USE THEM INSTEAD OF TAGON AND TAGOFF !!!
	# the variable "part" can have values :
	#	b : begin tag part
	#	m : middle tag part
	#	i : ignore tag part
	#	e : end tag part


	set part m
	set res {}
	foreach {k v i} $tmp {
	    switch -exact -- $k {
		text {
		    switch -exact -- $part {
			m {lappend res text $v}
			b {lappend res btag $v}
			e {lappend res etag $v}
			i {}
		    }
		}
		tagon {
		    switch -glob -- $v {
			btag* {set part i}
			etag* {set part i}
			default {}
		    }
		}
		tagoff {
		    switch -glob -- $v {
			btag* -
			etag* {
			    lappend res [lindex $v 0] [lindex $v 1]
			    set part m
			}
			default {}
		    }
		}
		default {
		    lappend res $k $v
		}
	    }
	}
 	return $res
    }

    # stext::makeTagID --
    #	get a tag id for a new tag. prepare all data for the tag.
    # Parameters:
    #	w: window name
    #	tag	: tag type
    #	parent	: parent tag
    #	options : various options for the tag
    # Options:
    # Results:
    # 	the new tag id
    # Side Effects:
    #	internal data about the new tag is set
    # Use:
    #	PRIVATE
    # Called by:

    proc makeTagID {w tag parent {options {}}} {
	variable VARS
	set num $VARS($w,maxtag)
	incr VARS($w,maxtag)
	set VARS($w,$num,type) $tag
	set VARS($w,$num,parent) $parent
	# Deal with options
	foreach {o v} $options {
	    setOption $w $num $o $v
	}
	# Deal with configure
	eval  [list $w.t tag configure  $num] [getTagConf [getDT $w] $tag conf]

	# get the parent tag configuration
	# get the modified configuration
	return $num
    }

    # stext::setOption --
    #	set the value of an option for a tag
    # Parameters:
    #	w: window name
    #	num	: tag id
    #	o	: option name
    #	v	: option value
    # Options:
    # Results:
    # Side Effects:
    #	the option value is set
    # Use:
    #	PUBLIC
    # Called by:
    
    proc setOption {w num o v} {
	variable VARS
	set VARS($w,$num,opt,$o) $v
    }

    # stext::getOption --
    #	get the value of an option for a tag
    # Parameters:
    #	w: window name
    #	num	: tag id
    #	o	: option name
    # Options:
    # Results:
    #	the the value of the option
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:
    
    proc getOption {w num o} {
	variable VARS
	return $VARS($w,$num,opt,$o)
    }

    # stext::getField --
    #	get working information attached to a tag
    #	the difference with option is that options are saved
    # Parameters:
    #	w: window name
    #	tagid	: tag id
    #	f	: field name
    # Options:
    # Results:
    #	the the value of the field
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:

    proc getField {w tagid f} {
	variable VARS
	return $VARS($w,$tagid,field,$f)
    }

    # stext::setField --
    #	set working information attached to a tag
    #	the difference with option is that options are saved
    # Parameters:
    #	w: window name
    #	tagid	: tag id
    #	f	: field name
    # Options:
    # Results:
    #	the the value of the field
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:

    proc setField {w tagid f v} {
	variable VARS
	set VARS($w,$tagid,field,$f) $v
    }

    # stext::createSubWindow --
    #	Prepares data for a subwindow tag.
    # Parameters:
    #	w: parent window name
    #	tagtype : tag type
    #	index	: position
    # Options:
    # Results:
    #	a tag id for the new window
    # Side Effects:
    #	tag id is incremented,
    #	a window name is built
    #	fields for the new tag are set
    # Use:
    #	PUBLIC
    # Called by:


    proc createSubWindow {w tagtype index {options {}}} {
	set parent [getTagID $w $index]
	set n [makeTagID $w $tagtype $parent $options]
	setField $w $n windowName $w.t.sub$n
	return $n
    }

    # stext::insertSubWindow --
    #	Inserts a window tag.
    #	the tag should have been declared with createSubWindow,
    #	and the window should have been effectively created.
    #	
    # Parameters:
    #	w: parent window name
    #	tagid : tag id
    # Options:
    # Results:
    # Side Effects:
    #	the window associated with tag id is created
    # Use:
    #	PUBLIC
    # Called by:

    proc insertSubWindow {w tagid} {
  	$w.t window create insert -window [getField $w $tagid windowName]
    }

    # stext::getTagConf --
    #	get the value of a tag configuration option
    # Parameters:
    #	doc : document type
    #	tag	: tag type
    #	o	: option name
    # Options:
    # Results:
    #	the the value of the option
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:
    
    proc getTagConf {doc tag o} {
	variable VARS
	return $VARS($doc,$tag,opt,-$o)
    }
    
    # stext::insertTagStart --
    #	inserts a tag start
    # Parameters:
    #	w: window name
    #	tagid	: tag id
    #	pos	: position of insertion
    # Options:
    # Results:
    # Side Effects:
    #	tag start is inserted
    # Use:
    #	PUBLIC
    # Called by:

    proc insertTagStart {w tagid pos} {
	set doc [getDT $w]
	set tag [getTagType $w $tagid]
	eval  [list $w.t tag configure [list btag $tagid]] [getTagConf $doc $tag beginconf]
	eval [getTagConf $doc $tag begin] [list $w [list $tagid [list btag $tagid]] $pos] 
    }

    # stext::insertTagEnd --
    #	inserts a tag end
    # Parameters:
    #	w: window name
    #	tagid	: tag id
    #	pos	: position of insertion
    # Options:
    # Results:
    # Side Effects:
    #	tag end is inserted
    # Use:
    #	PUBLIC
    # Called by:
    
    proc insertTagEnd {w tagid pos} {
	set doc [getDT $w]
	set tag [getTagType $w $tagid]
	eval [list $w.t tag configure [list etag $tagid]] [getTagConf $doc $tag endconf]
	eval [getTagConf $doc $tag end] [list $w [list $tagid [list etag $tagid]] $pos] 
    }

    # stext::insertText --
    #	insert a text at a given position
    # Parameters:
    #	w: window name
    #	text	: list representation of the text to insert
    # Options:
    # Results:
    # 	the new tag id
    # Side Effects:
    #	internal data about the new tag is set
    # Use:
    #	PRIVATE
    # Called by:

    #  proc insertText {w text pos} {
#  	set tagtype [lindex $text 0]
#  	set opts [lindex $text 1]
#  	set val [lindex $text 2]
#  	set parent [getParent $w $pos]
#  	if {[doctype::getKind [getDT $w]] == "pcdata"} {
#  	    set n [makeTagID $w $tagtype $parent $opts]
#  	    insertTagStart $w $n
#  	    insertTagEnd $w $n
#  	    moveIn $w $n
#  	    $w.t insert insert $val
#  	} else {
	    
#  	}
#      }
    

    # stext::getTagID --
    #	get the id of the tag at current position
    # Parameters:
    #	w: window name
    #	pos	: position
    # Options:
    # Results:
    # 	the tag id, if we are in a tag body.
    #	if we are in tag markup, returns -1
    # Side Effects:
    # Use:
    #	PRIVATE
    # Called by:
    
    proc getTagID {w pos} {
	set tags [$w.t tag names $pos]
	set res 0		;# if there is no tag, we are just in the document tag
	set hasmark 0		;# is there a boundary tag here ?
	# We know for sure 
	foreach t $tags {
	    switch -glob -- $t {
		sel {}
		btag* -
		etag* {
		    set hasmark 1
		    set boundmark $t
		}
		default {set res $t}
	    }
	}
	
	if {$res == 0} {
	    return 0
	}

	if {
	    [$w.t compare $pos >= [list btag $res].last] &&
	    [$w.t compare $pos <= [list etag $res].first]
	} {
	    puts "returning $res"
	    return $res
	} elseif {
	    $hasmark
	    && [$w.t compare $pos > $boundmark.first]
	    && [$w.t compare $pos < $boundmark.last]
	} {
	    puts "In boundary markup"
	    return -1
	} else {
	    puts "getting parent : [getParent $w $res]"
	    return [getParent $w $res]
	}
    }

    # stext::getTagType --
    #	get the type of a given tag
    # Parameters:
    #	w: window name
    #	tagid	: tag identifier
    # Options:
    # Results:
    # 	the tag type
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:
    
    proc getTagType {w tagid} {
	variable VARS
	return $VARS($w,$tagid,type)
    }

    # stext::getParent --
    #	get the id of a tag's parent
    # Parameters:
    #	w: window name
    #	tagid	: tag identifier
    # Options:
    # Results:
    # 	id of tag's parent
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:

    proc getParent {w tagid} {
	variable VARS
	return $VARS($w,$tagid,parent)	
    }

    # stext::moveIn --
    #	moves Inside a tagged text
    # Parameters:
    #	w: window name
    #	tagid	: tag identifier
    # Options:
    # Results:
    # Side Effects:
    #	The insertion mark is moved in the tagged text
    # Use:
    #	PUBLIC
    # Called by:

    proc moveIn {w tagid} {
	moveBefore $w $tagid
	$w.t mark set insert [list btag $tagid].last
    }

    # stext::moveBefore --
    #	moves before a tagged text
    # Parameters:
    #	w: window name
    #	tagid	: tag identifier
    # Options:
    # Results:
    # Side Effects:
    #	The insertion mark is moved just before the tagged text
    # Use:
    #	PUBLIC
    # Called by:

    proc moveBefore {w tagid} {
	$w.t mark set insert $tagid.first
    }

    # stext::moveAfter --
    #	moves after a tagged text
    # Parameters:
    #	w: window name
    #	tagid	: tag identifier
    # Options:
    # Results:
    # Side Effects:
    #	The insertion mark is moved just after the tagged text
    # Use:
    #	PUBLIC
    # Called by:


    proc moveAfter {w tagid} {
	$w.t mark set insert $tagid.last
    }

    # stext::textLeft --
    # Parameters:
    #	w: window name
    # Options:
    # Results:
    # Side Effects:
    #	The insertion mark is moved to the left at the first legal position.
    #	If it's not possible, don't move
    # Use:
    #	PUBLIC
    # Called by:

    proc textLeft {w} {
	# Recall position
	$w.t mark set tmp insert
	# Move
	$w.t mark set insert "insert - 1 chars"
	while {! [newDataIsLegal $w insert] && [$w.t compare insert > 1.0]} {
	    $w.t mark set insert "insert - 1 chars"
	}
	if {! [newDataIsLegal $w insert]} {
	    $w.t mark set insert tmp
	}
    }


    # stext::textRight --
    # Parameters:
    #	w: window name
    # Options:
    # Results:
    # Side Effects:
    #	The insertion mark is moved to the right at the first legal position
    # Use:
    #	PUBLIC
    # Called by:

    proc textRight {w} {
	# Record position
	$w.t mark set tmp insert
	# move
	$w.t mark set insert "insert + 1 chars"
	while {! [newDataIsLegal $w insert] && [$w.t compare insert < "end - 1 chars"] } {
	    $w.t mark set insert "insert + 1 chars"
	}
	if {! [newDataIsLegal $w insert]} {
	    $w.t mark set insert tmp
	}
    }

    # stext::nearestToMouse --
    #	moves insertion mark to the legal point nearest to the mouse
    # Parameters:
    #	w: window name
    # Options:
    # Results:
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:

    proc nearestToMouse {w x y} {
	puts hello
	$w.t mark set insert @$x,$y
	# First try to move right
	if { ! [newDataIsLegal $w insert]} {
	    textRight $w
	}
	# If it didn't solve the problem,  moves left.
	if { ! [newDataIsLegal $w insert]} {
	    textLeft $w
	}
	if { ! [newDataIsLegal $w insert]} {
	    error "No legal insertion point"
	}
	focus $w.t
	$w.t see insert
    }
    
    # stext::newDataIsLegal --
    # Parameters:
    #	w: window name
    #	p: position
    # Options:
    # Results:
    #	true if pos is a legal position
    #	A position is legal if some data can be inserted there. That is, if we are *inside*
    #	a choiceof tag.
    #	
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:
    
    proc newDataIsLegal {w pos} {
	set id [getTagID $w $pos]
	# If we are not in a tag body, we get -1 :
	if {$id == -1} {
	    return 0
	}
	set type [getTagType $w $id]
	set doc [getDT $w]
	# The tag should be an insertion tag.
	if {[string compare [doctype::getKind $doc $type] "choiceof" ] != 0} {
	    puts "data is NOT legal. Tag id is $id, tag is $type"
	    return 0
	}
	puts "data is legal"
	return 1
    }

    # stext::pcDataIsLegal --
    # Parameters:
    #	w: window name
    #	p: position
    # Options:
    # Results:
    #	true if pcdata is legal at position
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:

    proc pcDataIsLegal {w pos} {
	set id [getTagID $w $pos]
	# If we are not in a tag body, we get -1 :
	if {$id == -1} {
	    return 0
	}
	set type [getTagType $w $id]
	set doc [getDT $w]
	# The tag should be an insertion tag.
	return [doctype::hasData $doc $type]
    }
    
    # stext::getDT --
    #	returns the document type associated with a window
    # Parameters:
    #	w: window name
    # Options:
    # Results:
    # 	the doc. type
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:

    proc getDT {w} {
	variable VARS
	return $VARS($w,doctype)
    }

    # stext::tagPopup --
    #	build a popup menu fit for inserting new tags at current point.
    #	Of course, due to grammar restriction, this is only possible in "choiceof"
    #	tags.
    # Parameters:
    #	w: window name
    # Options:
    # Results:
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:
    
    proc tagPopup {w x y} {
	variable VARS

	set m $w.popup
	catch {destroy $m}
	menu $m

	# Get Doctype

	set doc [getDT $w]

	if {! [newDataIsLegal $w insert]} {
	    return
	}
	# get current tag
	set id [getTagID $w insert]
	set tag [getTagType $w $id]
	
	set choices [doctype::getChoice $doc $tag]
	if {[llength $choices] == 0} {
	    return
	}
 	foreach subtag $choices {
	    set command [getTagConf $doc $subtag create]
	    $m add command -label [ doctype::get $doc $subtag description ]\
			-command [list $command  $w $subtag insert]
	}
	catch {::tk_popup $m $x $y }
    }

    # stext::createFormat --
    #	creates a new format
    # Parameters:
    #	name	: format name
    # Options:
    # Results:
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:
    
    proc createFormat {name} {
	
    }

    # stext::declare --
    #	declare a tag realization in a format
    # Parameters:
    #	format	: format name
    #	tag	: tag name
    # Options:
    #	-conf		: fonts, colour, etc... configuration passed to text widget tag configuration
    #	-begin		: procedure which will build a heading for the tag. it will be evaluated,
    #		  	 so values like "plainInsert {translation :}" are ok.
    #		  	 the function will receive three additional arguments, which are window, tagId and index.
    #		  	 see plainInsert for an example. The result SHOULD NOT BE EMPTY (-elide should be used instead).
    #	-end		: procedure which will build an ending for the tag. see -begin
    #	-beginconf 	: configuration for the start tag. e.g. "-background red" 
    #	-endconf 	: configuration for the end tag.
    #	-init		: initialization data.
    #	-create 	: function called to insert a new tag
    #				arguments are :
    #	-load		: load function
    #	-save		: save function
    #			  the function will receive the text as argument, save it, and return
    #			  a replacement text, which will be saved.

    # Results:
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:
    
    proc declare {doctype tag args} {
	variable VARS
	myopts::myopts opts { 
	    -conf {}
	    -begin {plainInsert {<}}
	    -end   {plainInsert {>}}
	    -beginconf {-background red -elide 1}
	    -endconf {-background red -elide 1}
	    -init {}
	    -create {stext::simpleInsertTag}
	    -load {}
	    -save {stext::simpleSave}
	} $args
        foreach {o v} [array get opts] {
	    set VARS($doctype,$tag,opt,$o) $v
	}
    }

    # stext::plainInsert --
    #	simple utility procedure. The purpose of this function is to be
    # 	used as argument of options, like -begin (see example).
    # Parameters:
    # Options:
    # Results:
    # Side Effects:
    # Use:
    #	PUBLIC
    # Called by:

    proc plainInsert {txt w tagsid index} {
	puts "$w.t insert $index $txt $tagsid"
	$w.t insert $index $txt $tagsid
    }

    # stext::simpleInsertTag --
    # 	inserts a tag at current point. If needed, the tag's minimal
    # 	content is inserted too.
    #
    #	This is the procedure which is called when adding a new tag from the
    #	menu.
    #
    # Parameters:
    #	w	: window
    #	tag	: tag type to insert
    #	index	: where to insert (or should we use "insert" ?)
    # Options:
    # Results:
    #	The tag is inserted. 
    #	The insertion mark is moved into the first legal position in the tag
    # Side Effects:
    # Use:
    # Called by:
    
    proc simpleInsertTag {w tag index} {
	set parent [getTagID $w $index]
	set n [makeTagID $w $tag $parent]
	insertTagStart $w $n $index
	insertTagEnd $w $n $index
	moveIn $w $n
	insertNewTagContent $w $tag insert
	# Moves to the first legal position
	moveBefore $w $n
	textRight $w
	return $n
    }

    # stext::insertNewTagContent --
    #	insert the default content of a tag at a given point.
    #	that is, do nothing except for "seq" tags.
    # Parameters:
    # Options:
    # Results:
    # Side Effects:
    # Use:
    # Called by:

    proc insertNewTagContent {w tag index} {
	set doc [getDT $w]
	set k [doctype::getKind $doc $tag]
	if {[string compare $k "seq"] == 0} {
	    foreach i [doctype::getChoice $doc $tag] {
		set n [simpleInsertTag $w $i insert]
		moveAfter $w $n
	    }
	}
    }

    # stext::simpleSave --
    #	Default save function for a tag.
    #		Save functions get the tag's content as argument,
    #		and their responsability is to send back a list corresponding to
    #		what should actually be saved.
    # Parameters:
    # Options:
    # Results:
    # Side Effects:
    # Use:
    # Called by:
    
    proc simpleSave {w txt} {
	foreach {tagid content} $txt {}
	set res {}
	set doc [getDT $w]
	set tag [getTagType $w $tagid]
	foreach t $content {
	    if [isData $t] {
		lappend res $t
	    } else {
		lappend res [[getTagConf $doc $tag save] $t]
	    }
	}
	return [list $tag [getTagOptions $w $tagid] $content]
    }
    
    # Two other saving methods :
    # container tags
    # stand-alone subtags
    # stext::baseSave --
    # 		A tag with this saving system will be saved in the following way :
    #		The tag will be saved as an entry in the database (tag attribute "id")
    #		The function will then return the tag Id.
    #		All subtags will gain a "position" attribute (pos), which defines their position
    #		Of course, 

    
    # Parameters:
    # Options:
    # Results:
    # Side Effects:
    # Use:
    # Called by:
    

    # stext::insertNewTag --
    #	insert a new tag
    # Parameters:
    # Options:
    # Results:
    # Side Effects:
    # Use:
    # Called by:

    #proc insertNewTag {w tag} {
	#	set n [simpleInsertTag $w $tag insert]
	#	moveAfter $w $n
	#   }
    
    # stext::keyInsert --
    #	insert a key at current point
    # Parameters:
    # Options:
    # Results:
    # Side Effects:
    # Use:
    # Called by:

    proc keyInsert {w key} {
	#textInsert $win [keyvalue $win $k]	
	textInsert $w $key
    }

    # stext::textInsert --
    #	insert pcdata at current point
    # Parameters:
    # Options:
    # Results:
    # Side Effects:
    # Use:
    # Called by:

    proc textInsert {w txt} {
	if [pcDataIsLegal $w insert] {
	    $w.t insert insert $txt [getTagID $w insert]
	    $w.t see insert
	}
	# upper_tag $win
    }

    # stext::call --
    #	bound keys and the like take the $w.t window as argument,
    #	but normal functions take $w.
    #	call is an utility function to fix this.
    #	"call function $w.t arg1.... "
    #	will in fact call : function $w arg1...
    # Parameters:
    #	func	: function to call
    #	win	: text window
    #	args	: other arguments to pass to the function
    # Options:
    # Results:
    # Side Effects:
    #	function is called
    # Use:
    #	HALF-PUBLIC :-
    # Called by:
    
    proc call {func win args} {
	# A REMETTRE EN PLACE DES QUE J'AI UNE VERSION RECENTE DE TCL SOUS LA MAIN !!!
	set w [string range $win 0 "end-2"]
	# set l [string length $win]
	# incr l -3
	# set w [string range $win 0 $l]
	eval $func $w $args
    }
}

# Saving data :
# call a function for each subtag.
# the function should send back *what* should be saved
# Possibility : 
# don't use dump that way.
# 

    # declare_tag --
    # procedure for declaring a TAG in a document style.
    #
    #
    # declare_tags

    # -grammar : regexp showing how tags should be entered.

    # -begin / -end : begin and end marks. 

    # -defaultval : default value
    
    # -conf / -bconf / -econf : the way configuration should be handled

    # -proc_conf / -proc_bconf / -proc_econf : idem, but a procedure is provided to 
    #                                          compute the presentation.
   
    # -savewindow / -loadwindow : procedures to save/load a window tag.
    # the save proc takes : the widget and the window as arguments
    # the load proc arguments should be {w {txt ""}} or something like that

    # -save / -load : procedures to save/load a normal tag.
    # args are
    # save : {w txt}
    # load : {w txt}. The txt is the the tag value in the source. Now, if 
    #   -load is specified, the tag value probably isn't the "real" tag value to insert.

    # -index : should tag value be indexed ?
    
    # -normalize : normalization function used as filter for indexed tags

    # -description : full text description of the tag meaning

    # -ext_save : save function for window tags in external format (latex, HTML)

    # -ext_begin : string added in front of a tag when saving in external format.

    # -ext_end : idem, for the ending.

    # -insert_at_end : should the insertion point be placed at the end of the tag ?

    # -bindings 
 
    # -keybind

    # internal representation
    # VARS(<doctype>,<tag>,...)
    # 
    
#      proc DeclareTag {doctype tagname args} {
#  	variable VARS
#  	myopts opts {
#  	    -grammar {}
#  	} $args
#  	set VARS($doctype,$tagname,grammar) $opts(-grammar)
#      }

# About Tags : 
# a tag is identified by a number.
# hence, we have a table of tags in 
# VARS($w,$NUM,...)
# among saved information :
# - list of options ?
# - value for each option :
# 	VARS($w,$NUM,opt,$OPTNAME)
# - parent (for stext) : tag number of parent tag.
# - tag type of tag
# - tag id of tag (it's an option)

# Inserting a tag :
# set type $VARS($w,$tag,type)
# set parent $VARS($w,parent)
# set pt $VARS($w,$parent,type)
# if doctype::isPCData $type and doctype::getkind $pt == choiceof
#    then
#	break current tag in two tags :
#       (let's make this a proc breakTag)
#	create a new tag, duplicating the old one.
#	tag remove the tag between insert and $tag.last
#       insert tag ending and tag start at insert point
#	tag add the new tag value between the new start tag and the old end tag
#
#	Now we can insert the element
# else if the tag is legal, insert the element.




################################################################################
#
# Special bindings for Text widget containing hierarchic tags.
#

bind HierarchicText <Control-Tab> {
    focus [tk_focusNext %W]
}

bind HierarchicText <Control-Shift-Tab> {
    focus [tk_focusPrev %W]
}

bind HierarchicText <Delete> {stext::call textdelete %W}

bind HierarchicText <BackSpace> {stext::call textbackspace %W}

bind HierarchicText <KeyPress> {stext::call keyInsert %W %A}

bind HierarchicText <Return> {stext::call keyInsert  %W "\n"}

bind HierarchicText <Button-3> {
    stext::call tagPopup %W %X %Y
}

bind HierarchicText <Menu> {
    stext::call tagPopup %W %X %Y
}

bind HierarchicText <Key-F2> {
    stext::call tagPopup %W %X %Y
}

bind HierarchicText <Button-1> {
    stext::call nearestToMouse %W %x %y
}

bind HierarchicText <Left> {stext::call textLeft %W}

bind HierarchicText <Right> {stext::call textRight %W}

bind HierarchicText <Up> {stext::call textup %W}

bind HierarchicText <Down> {stext::textdown %W}

bind HierarchicText <<Paste>> {stext::winpastetext %W}

bind HierarchicText <Control-s> {stext::call toggleElide %W}

bind HierarchicText <Key-Escape><Key-d> {puts here; stext::call makeList %W}

bind HierarchicText <Key-Escape> {set tmp tmp}



